from typing import Union

import altair as alt
import numpy as np
import pandas as pd
import xarray as xr


def _standardize_data(source: Union[pd.DataFrame, xr.Dataset],
                      z: str = 'time') -> pd.DataFrame:
    """Convert Dataset to Dataframe and add required columns for plotting."""
    if isinstance(source, xr.Dataset):
        source = source.to_dataframe().reset_index()

    if 'run' not in source:
        source['run'] = 0

    if 'slider' not in source:
        _, idx = np.unique(source[z], return_inverse=True)
        source['slider'] = idx

    return source


def alt_line_chart(source: Union[pd.DataFrame, xr.Dataset],
                   *,
                   x: str,
                   y: str,
                   z: str = 'time',
                   std: bool = False) -> alt.Chart:
    """Generate an altair line chart from a dataframe.

    Parameters
    ----------
    source : pd.DataFrame
        Input dataframe
    x : str
        X-value to plot, corresponds to a column in the source data
    y : str
        Y-value to plot, corresponds to a column in the source data
    z : str
        Slider variable (time), corresponds to a column in the source data

    std : bool
        Plot the error bound from {x}_error_upper in the plot as well

    Returns
    -------
    alt.Chart
        Return an altair chart.
    """
    source = _standardize_data(source, z=z)
    max_y = source[y].max()

    if std:
        source[y + '_upper'] = source[y] + source[y + '_error_upper']
        source[y + '_lower'] = source[y] - source[y + '_error_upper']
        max_y = source[y + '_upper'].max()

    max_slider = source['slider'].max()

    if std:
        band = alt.Chart(source).mark_area(opacity=0.3).encode(
            x=f'{x}:Q',
            y=alt.Y(f'{y}_upper:Q', title=y),
            y2=alt.Y2(f'{y}_lower:Q', title=y),
            color=alt.Color('run:N'),
        )

    line = alt.Chart(source).mark_line().encode(
        x=f'{x}:Q',
        y=alt.Y(
            f'{y}:Q',
            scale=alt.Scale(domain=(0, max_y)),
            axis=alt.Axis(format='.4~g'),
        ),
        color=alt.Color('run:N'),
        tooltip='run',
    )

    ref = alt.Chart(source).mark_line(strokeDash=[5, 5]).encode(
        x=f'{x}:Q', y=f'{y}:Q', color=alt.Color('run:N'), tooltip='run')

    if max_slider != 0:
        slider = alt.binding_range(name='Time index',
                                   min=0,
                                   max=max_slider,
                                   step=1)
        select_step = alt.selection_single(name=z,
                                           fields=['slider'],
                                           bind=slider,
                                           init={'slider': 0})
        line = line.add_selection(select_step).transform_filter(
            select_step).interactive()
        if std:
            band = band.transform_filter(select_step).interactive()

        first_run = source.iloc[0].run
        slider = alt.binding_range(name='Reference time index',
                                   min=0,
                                   max=max_slider,
                                   step=1)
        select_step = alt.selection_single(name='reference',
                                           fields=['slider'],
                                           bind=slider,
                                           init={'slider': 0})

        ref = ref.add_selection(select_step).transform_filter(
            select_step).transform_filter(
                alt.datum.run == first_run).interactive()

    if std:
        return line + ref + band
    else:
        return line + ref


def alt_errorband_chart(source: Union[pd.DataFrame, xr.Dataset],
                        *,
                        x: str,
                        y: str,
                        z: str = 'time') -> alt.Chart:
    """Generate an altair errorband plot from a dataframe.

    Parameters
    ----------
    source : pd.DataFrame
        Input dataframe
    x : str
        X-value to plot, corresponds to a column in the source data
    y : str
        Y-value to plot, corresponds to a column in the source data
    z : str
        Slider variable (time), corresponds to a column in the source data

    Returns
    -------
    alt.Chart
        Return an altair chart.
    """
    source = _standardize_data(source, z=z)

    max_y = source[y].max()
    max_slider = source['slider'].max()

    slider = alt.binding_range(min=0, max=max_slider, step=1)
    select_step = alt.selection_single(name=z,
                                       fields=['slider'],
                                       bind=slider,
                                       init={'slider': 0})

    line = alt.Chart(source).mark_line().encode(
        x=f'{x}:Q',
        y=alt.Y(
            f'mean({y}):Q',
            scale=alt.Scale(domain=(0, max_y)),
            axis=alt.Axis(format='.4~g'),
        ),
        color=alt.Color('tstep:N'),
    ).add_selection(select_step).transform_filter(select_step).interactive()

    # altair-viz.github.io/user_guide/generated/core/altair.ErrorBandDef
    band = alt.Chart(source).mark_errorband(
        extent='stdev', interpolate='linear').encode(
            x=f'{x}:Q',
            y=f'{y}:Q',
            color=alt.Color('tstep:N'),
        ).add_selection(select_step).transform_filter(
            select_step).interactive()

    slider = alt.binding_range(name='Reference time index',
                               min=0,
                               max=max_slider,
                               step=1)
    select_step = alt.selection_single(name='reference',
                                       fields=['slider'],
                                       bind=slider,
                                       init={'slider': 0})

    ref = alt.Chart(source).mark_line(strokeDash=[5, 5]).encode(
        x=f'{x}:Q',
        y=f'mean({y}):Q',
        color=alt.Color('tstep:N'),
    ).add_selection(select_step).transform_filter(select_step).interactive()

    return line + band + ref



from typing import Any, Optional

from pydantic import Field

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._jetto import JettoVar


class JettoVariableModel(BaseModel):
    """Variable for describing variables specific to Jetto, The lookup table
    can be defined as a JettoVar under the lookup key."""

    type: str = Field('jetto-variable',
                      description='Discriminator for the variable type.')

    name: str = Field(description=f("""
        Name of the variable.
        Used for the lookup table to find actual fields.
        """))
    lookup: Optional[JettoVar] = Field(description=f("""
    Description of the fields that have to be updated for a Jetto Variable
    """))


class IDSPath(BaseModel):
    ids: str = Field(description='Root IDS name.')
    path: str = Field(description=f("""
        Path to the data within the IDS.
        The fields are separated by forward slashes (`\\`).
    """))


class IDSVariableModel(IDSPath):
    """Variable for describing data within a IMAS database.

    The variable can be given a name, which will be used in the rest of the config
    to reference the variable. It will also be used as the column labels or
    on plots.

    The dimensions for each variable must be specified. This ensures the the data
    will be self-consistent. For example for 1D data, you can use `[x]` and for 2D data,
    `[x, y]`.

    The IDS path may contain indices. You can point to a single index, by simply giving the
    complete path (i.e. `profiles_1d/0/t_i_ave` for the 0th time slice).
    To retrieve all time slices, you can use `profiles_1d/*/t_i_ave`.
    """
    type: str = Field('IDS-variable',
                      description='discriminator for the variable type')

    name: str = Field(description=f("""
        Name of the variable.
        This will be used to reference this variable.
    """))
    ids: str = Field(description='Root IDS name.')
    path: str = Field(description=f("""
        Path to the data within the IDS.
        The fields are separated by forward slashes (`/`).
    """))
    dims: list[str] = Field(description=f("""
        Give the dimensions of the data,
        i.e. [x] for 1D, or [x, y] for 2D data.
    """))


class Condition(BaseModel):
    """Only accept value if this condition returns `True`.

    Equivalent to: `operator(X, *args)`, where `X` is the value
    retrieved from the IDS paths.
    """
    operator: str = Field(
        'Name of a function in the `operator` module in the standard library.')
    args: list[Any] = Field('Arguments to pass to the operator.')

    def __call__(self, value: Any):
        import operator
        test = getattr(operator, self.operator)
        return test(value, *self.args)


class IDS2JettoVariableModel(BaseModel):
    """Variable for describing the relation between IDS data and jetto
    variables.

    The variable can be given a name, which can be used in the config
    template to reference the variable. It will also be used as the
    column labels or on plots.
    """
    name: str = Field(description=f("""
        Name of the variable.
        This will be used to reference this variable.
    """))
    type: str = Field('IDS2jetto-variable',
                      description='discriminator for the variable type')
    paths: list[IDSPath] = Field(description=f("""
        Search these variables in the given order until a match with the conditions
        defined below is found.
    """))
    accept_if: Optional[list[Condition]] = Field([],
                                                 description=f("""
        Accept variable if it matches the given conditions. This can be used
        to filter undefined values (i.e. set to 0 or some very small or large number).
        """))
    default: Optional[float] = Field(description=f("""
        Default value if no match is found. Set to None to raise an exeption instead."""
                                                   ))



from __future__ import annotations

import logging
from pathlib import Path

import click
import xarray as xr

from ._plot_utils import alt_line_chart
from .config import var_lookup
from .ids import ImasHandle, rebase_all_coords
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def plot(*, var_names, handles, input_files, extensions, errorbars, **kwargs):
    handle_lst = []

    for n, imas_str in enumerate(handles):
        handle = ImasHandle.from_string(imas_str)
        handle_lst.append(handle)

    handles = {n: handle for n, handle in enumerate(handle_lst)}

    for input_file in input_files:
        handles.update(read_imas_handles_from_file(input_file))

    if len(handles) == 0 or len(var_names) == 0:
        raise SystemExit('No data to show.')

    for n, variable in enumerate(var_lookup[var_name]
                                 for var_name in var_names):
        data_var = variable.name
        time_var = variable.dims[0]
        grid_var = variable.dims[1]

        grid_var_norm = var_lookup.normalize(grid_var)

        variables = [data_var, grid_var, time_var]

        if errorbars:
            data_var_std = variable.copy()
            data_var_std.path += '_error_upper'
            data_var_std.name += '_error_upper'

            variables.append(data_var_std)

        datasets = []
        for handle in handles.values():
            ds = handle.get_variables(variables=variables)
            datasets.append(ds)

        datasets = rebase_all_coords(datasets, datasets[0])
        dataset = xr.concat(datasets, 'run')

        chart = alt_line_chart(dataset,
                               x=grid_var_norm,
                               y=data_var,
                               z=time_var,
                               std=errorbars)

        click.echo('You can now view your plot in your browser:')
        click.echo('')

        click.secho(f'  {grid_var_norm} vs. {data_var}:\n', fg='green')

        for extension in extensions:

            outfile = Path(f'chart_{grid_var_norm}-{data_var}.{extension}')
            click.secho(f'    file:///{outfile.absolute()}', bold=True)

            chart.save(outfile, scale_factor=2.0)

        click.echo('')



from __future__ import annotations

import os
from collections import defaultdict
from contextlib import contextmanager
from itertools import filterfalse, tee
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Hashable, Iterable

from ._types import PathLike
from .schema.runs import Runs

if TYPE_CHECKING:
    from .ids import ImasHandle


def no_op(*args, **kwargs):
    """Do nothing."""
    pass


@contextmanager
def work_directory(path: PathLike):
    """Changes working directory and returns to previous on exit.

    Parameters
    ----------
    path : PathLike
        Temporarily change to this directory.
    """
    prev_cwd = Path.cwd().resolve()
    try:
        os.chdir(path)
        yield
    finally:  # In any case, no matter what happens, go back eventually
        os.chdir(prev_cwd)


def read_imas_handles_from_file(inp: PathLike, ) -> dict[str, ImasHandle]:
    """Read a collection of imas paths from a file.

    Input can be a `Runs.yaml` file `data.csv` file.

    The CSV file must have contain at least 5 columns, including: `user`,
    `db`, `shot`, and `run`. The first column is used as the index.
    The index can be any string or number, as long as it uniquely
    identifies the row.

    Parameters
    ----------
    inp : PathLike
        Name of the file to read.
    as_dataframe : bool, optional
        Return imas handles in a dataframe instead.

    Returns
    -------
    Union[dict[str, ImasHandle], 'pd.DataFrame']
        Returns a dict with the Imas handles.

    Raises
    ------
    ValueError
        When the file cannot be opened.
    """
    import csv

    from .ids import ImasHandle

    inp = Path(inp)

    if inp.suffix == '.csv':
        handles = {}
        with open(inp) as f:
            reader = csv.DictReader(f)

            index_col = reader.fieldnames[0]  # type: ignore

            for row in reader:
                index = row.pop(index_col)
                handles[index] = ImasHandle(**row)

    elif inp.name == 'runs.yaml':
        runs = Runs.parse_file(inp)
        handles = {
            str(run.dirname): ImasHandle.parse_obj(run.data_out)
            for run in runs
        }

    else:
        raise ValueError(f'Cannot open file: {inp}')

    return handles


def groupby(iterable: Iterable,
            keyfunc: Callable) -> dict[Hashable, list[Any]]:
    """Group iterable by key function. The items are grouped by the value that
    is returned by the `keyfunc`

    Parameters
    ----------
    iterable : list, tuple or iterable
        List of items to group
    keyfunc : callable
        Used to determine the group of each item. These become the keys
        of the returned dictionary

    Returns
    -------
    grouped : dict
        Returns a dictionary with the grouped values.
    """
    grouped = defaultdict(list)
    for item in iterable:
        key = keyfunc(item)
        grouped[key].append(item)

    return grouped


def partition(pred: Callable, iterable: Iterable) -> tuple[Iterable, Iterable]:
    """Use a predicate to partition entries into false entries and true
    entries.

    partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9

    From: https://docs.python.org/3/library/itertools.html

    Parameters
    ----------
    pred : Callable
        Filter function called on every entry in the iterable to determine
        whether to put it in the false or true bin
    iterable : Iterable
        List of items to partition

    Returns
    -------
    tuple[Iterable, Iterable]
        Two sequences with false and true entries, respectively.
    """
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)



from __future__ import annotations

import re
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Sequence

import numpy as np

from duqtools.config import lookup_vars

from ..schema import IDSVariableModel
from ._copy import add_provenance_info

if TYPE_CHECKING:
    import xarray as xr

    from ._handle import ImasHandle

INDEX_STR = '*'


class EmptyVarError(Exception):
    ...


def insert_re_caret_dollar(string: str) -> str:
    """Insert regex start (^) / end ($) of line matching characters."""
    if not string.startswith('^'):
        string = f'^{string}'
    if not string.endswith('$'):
        string = f'{string}$'
    return string


def replace_index_str(string: str) -> str:
    """Replaces template string with regex digit matching."""
    return string.replace(INDEX_STR, r'(\d+)')


class IDSMapping(Mapping):

    def __init__(self, ids):
        """Map the IMASDB object.

        Empty arrays are excluded from the mapping.
        You can still get/set these keys directly,
        but `key in map` returns `False` if `map['key']` is an empty array.

        Parameters
        ----------
        ids :
            IMAS DB entry for the IDS.

        Attributes
        ----------
        exclude_empty : bool
        """
        self._ids = ids

        # All available data fields are stored in this set.
        self._keys: set[str] = set()
        self._paths: dict[str, Any] = {}

        self.dive(ids, [])

    def __repr__(self):
        s = f'{self.__class__.__name__}(\n'
        for key in self._paths:
            s += f'  {key} = ...\n'
        s += ')\n'

        return s

    @staticmethod
    def _getattr(pointer: Any, attr: str) -> Any:
        """Like `getattr`, but will return the index if `attr` is an int."""
        try:
            i = int(attr)
        except (ValueError, TypeError):
            ret = getattr(pointer, attr)
        else:
            ret = pointer[i]  # type: ignore

        return ret

    def _deconstruct_key(self, key: str) -> tuple[Any, str]:
        """Break down key and return pointer + attr."""
        *parts, attr = key.split('/')

        pointer = self._ids

        for part in parts:
            pointer = self._getattr(pointer, part)

        return pointer, attr

    def __getitem__(self, key: str) -> Any:

        try:
            pointer, attr = self._deconstruct_key(key)
            ret = self._getattr(pointer, attr)
        except AttributeError as err:
            raise KeyError(key) from err
        except IndexError as err:
            raise KeyError(key) from err

        return ret

    def __setitem__(self, key: str, value: np.ndarray):

        try:
            pointer, attr = self._deconstruct_key(key)
            getattr(pointer, attr)
        except AttributeError as err:
            raise KeyError(key) from err
        except IndexError as err:
            raise KeyError(key) from err
        else:
            setattr(pointer, attr, value)

    def __iter__(self):
        yield from self._keys

    def __len__(self):
        return len(self._keys)

    def __contains__(self, key):
        return key in self._keys

    @staticmethod
    def _path_at_index(variable: str | IDSVariableModel,
                       index: int | Sequence[int]):
        path = variable.path if isinstance(variable,
                                           IDSVariableModel) else variable

        if isinstance(index, int):
            index = (index, )

        for i in index:
            path = path.replace('*', str(i), 1)

        return path

    def get_at_index(self, variable: str | IDSVariableModel,
                     index: int | Sequence[int], **kwargs) -> Any:
        """Grab key with index replacement.

        Example: `IDSMapping.get_at_index(var, index=0)`
        """
        path = self._path_at_index(variable, index)
        return self[path]

    def set_at_index(self, variable: str | IDSVariableModel,
                     index: int | Sequence[int], value: Any, **kwargs):
        """Grab key with index replacement.

        Example: `IDSMapping.set_at_index(var, value, index=0)`
        """
        path = self._path_at_index(variable, index)
        self[path] = value

    def length_of_key(self, key: str):
        """length_of_key gives you the number of entries of a (partial) ids
        path, or None if the length does not exist.

        Note: this is different then the length of an IDSMapping, which is defined
        as the number of keys

        Note: calling `len(map[key])` works as well

        ## Example:


        ```python
        map.length_of_key('1d_profiles')
        ```

        Parameters
        ----------
        key : str
            key
        """
        try:
            return len(self[key])
        except Exception:
            pass

    def sync(self, target: ImasHandle):
        """Synchronize updated data back to IMAS db entry.

        Shortcut for 'put' command.

        Parameters
        ----------
        target : ImasHandle
            Points to an IMAS db entry of where the data should be written.
        """

        add_provenance_info(handle=target)

        with target.open() as db_entry:
            self._ids.put(db_entry=db_entry)

    def dive(self, val, path: list):
        """Recursively find the data fields.

        Parameters
        ----------
        val :
            Current nested object being evaluated
        path : list
            Current path
        """

        if isinstance(val, str):
            return

        if hasattr(val,
                   '__getitem__') and not isinstance(val,
                                                     (np.ndarray, np.generic)):
            for i in range(len(val)):
                item = val[i]
                self.dive(item, path + [str(i)])
            return

        if hasattr(val, '__dict__'):
            for key, item in val.__dict__.items():
                self.dive(item, path + [key])
            return

        if not isinstance(val, (np.ndarray, np.generic)):
            return

        if val.size == 0:
            return

        # We made it here, the value can be stored
        str_path = '/'.join(path)
        self._keys.add(str_path)

        cur = self._paths
        for part in path[:-1]:
            cur.setdefault(part, {})
            cur = cur[part]
        cur[path[-1]] = str_path

    def findall(self, pattern: str) -> dict[str, Any]:
        """Find keys matching regex pattern.

        Parameters
        ----------
        pattern : str
            Regex pattern

        Returns
        -------
        dict
            New dict with all matching key/value pairs.
        """
        pattern = insert_re_caret_dollar(pattern)
        pattern = replace_index_str(pattern)

        pat = re.compile(pattern)

        return {key: self[key] for key in self._keys if pat.match(key)}

    def find_by_group(self, pattern: str) -> dict[tuple | str, Any]:
        """Find keys matching regex pattern by group.

        The dict key is defined by `match.groups()`.
        Dict entries will be overwritten if the groups are not unique.

        Parameters
        ----------
        pattern : str
            Regex pattern (must contain groups)

        Returns
        -------
        dict
            New dict with all matching key/value pairs.
        """
        pattern = insert_re_caret_dollar(pattern)
        pattern = replace_index_str(pattern)

        pat = re.compile(pattern)

        new = {}
        for key in self._keys:
            m = pat.match(key)
            if m:
                groups = m.groups()
                idx = groups[0] if len(groups) == 1 else groups
                new[idx] = self[key]

        return new

    def _read_array_from_parts(self, *parts: str):
        arr = []
        root, sub, *remaining = parts
        nodes = self[root]

        for index in range(len(nodes)):
            path = f'{root}/{index}/{sub}'

            if remaining:
                sub_arr = self._read_array_from_parts(path, *remaining)
            else:
                sub_arr = self[path]

            arr.append(sub_arr)

        return arr

    def to_xarray(
        self,
        variables: Sequence[str | IDSVariableModel],
        empty_var_ok: bool = False,
        **kwargs,
    ) -> xr.Dataset:
        """Return dataset for given variables.

        Parameters
        ----------
        variables : Sequence[str | IDSVariableModel]]
            Dictionary of data variables
        empty_var_ok : bool
            If True, silently skip data that are missing from the mapping.
            If False (default), raise an error when data that are missing
            from the dataset are requested.

        Returns
        -------
        ds : xr.Dataset
            Return query as Dataset
        """

        def _contains_empty(arr):
            if isinstance(arr, list):
                if len(arr) == 0:
                    return True
                return any(_contains_empty(sub_arr) for sub_arr in arr)
            elif isinstance(arr, np.ndarray):
                return arr.size == 0
            elif isinstance(arr, (float, int)):
                return False
            else:
                raise ValueError(
                    f"Don't know how to deal with: {var.name}: {arr}")

        import xarray as xr

        xr_data_vars: dict[str, tuple[list[str], np.ndarray]] = {}

        variables = lookup_vars(variables)

        for var in variables:
            parts = var.path.split('/*/')

            if len(parts) == 1:
                xr_data_vars[var.name] = (var.dims, self[var.path])
                continue

            arr = self._read_array_from_parts(*parts)

            if _contains_empty(arr):
                if empty_var_ok:
                    continue
                else:
                    raise EmptyVarError(
                        f'Variable {var.name!r} contains empty data.')
            xr_data_vars[var.name] = ([*var.dims], arr)

        ds = xr.Dataset(data_vars=xr_data_vars)  # type: ignore

        return ds

    def _write_array_in_parts(self, data, *parts: str) -> None:
        """_write_array_in_parts.

        inner function that determines the path and writes back the data
        """
        if len(parts) < 2:
            # Write back
            path, = parts
            self[path] = data
            return

        root, sub, *remaining = parts
        nodes = self[root]
        for index in range(len(nodes)):
            path = f'{root}/{index}/{sub}'
            self._write_array_in_parts(data[index], path, *remaining)

    def write_array_in_parts(self, variable_path: str,
                             data: xr.DataArray) -> None:
        """write_back data, give the data, and the variable path, where `*`
        denotes the dimensions. This function will figure out how to write it
        back to the IDS.

        Parameters
        ----------
        variable_path : str
            path of the variable in the IDS, something like 'profiles_1d/*/zeff'
        data : xr.DataArray
            data of the variable, in the correct dimensions (every star in the
            `variable_path` is a dimension in this array)

        Returns
        -------
        None
        """
        parts = variable_path.split('/*/')
        self._write_array_in_parts(data.data, *parts)



from pathlib import Path

from ..config import cfg
from ..create import create as duqtools_create
from ..utils import work_directory


def create(**kwargs):
    cwd = Path.cwd()

    config_files = cwd.glob('**/duqtools.yaml')

    for config_file in config_files:
        cfg.parse_file(config_file)

        config_dir = config_file.parent

        with work_directory(config_dir):
            duqtools_create(config=config_file, **kwargs)



import logging
import shutil
import warnings
from pathlib import Path
from typing import Any, Sequence

import pandas as pd

from .apply_model import apply_model
from .cleanup import remove_run
from .config import cfg
from .ids import ImasHandle
from .matrix_samplers import get_matrix_sampler
from .models import Locations
from .operations import add_to_op_queue, op_queue
from .schema.runs import Run, Runs
from .system import get_system

logger = logging.getLogger(__name__)

RUN_PREFIX = 'run_'


class CreateError(Exception):
    ...


class CreateManager:
    """Docstring for CreateManager."""

    def __init__(self):
        self.options = cfg.create

        if not self.options:
            logger.warning('No create options specified.')
            raise CreateError('No create options specified in config.')

        self.template_drc = self.options.template
        self.system = get_system()
        self.runs_dir = self.system.get_runs_dir()
        self.source = self._get_source_handle()

        locations = Locations()
        self.runs_yaml = locations.runs_yaml
        self.data_csv = locations.data_csv

    def _get_source_handle(self) -> ImasHandle:
        if not self.options.template_data:
            source = self.system.imas_from_path(self.template_drc)
        else:
            source = ImasHandle.parse_obj(self.options.template_data)

        logger.info('Source data: %s', source)

        return source

    def generate_ops_list(self) -> list[Any]:
        """Generate set of operations for a run."""
        dimensions = self.options.dimensions
        matrix_sampler = get_matrix_sampler(self.options.sampler.method)
        matrix = tuple(model.expand() for model in dimensions)
        ops_list = matrix_sampler(*matrix, **dict(self.options.sampler))

        return ops_list

    def make_run_models(self, ops_list: Sequence[Any]) -> list[Run]:
        """Take list of operations and create run models."""
        run_models = []

        for i, operations in enumerate(ops_list):
            dirname = self.runs_dir / f'{RUN_PREFIX}{i:04d}'

            data_in = self.system.get_data_in_handle(
                dirname=dirname,
                seq_number=i,
                source=self.source,
                options=self.options.data,
            )

            data_out = self.system.get_data_out_handle(
                dirname=dirname,
                seq_number=i,
                source=self.source,
                options=self.options.data,
            )

            model = Run(dirname=dirname.resolve(),
                        shortname=dirname.name,
                        data_in=data_in,
                        data_out=data_out,
                        operations=operations)

            run_models.append(model)

        return run_models

    def runs_yaml_exists(self) -> bool:
        """Check if runs.yaml exists."""
        if self.runs_yaml.exists():
            op_queue.add_no_op(description='Not creating runs.yaml',
                               extra_description=f'{self.runs_yaml} exists')
            return True
        return False

    def data_locations_exist(self, models: Sequence[Run]) -> bool:
        """Check IDS coordinates and raise if any exist."""
        any_exists = False

        for model in models:
            if ImasHandle.parse_obj(model.data_in).exists():
                logger.info('Target %s already exists', model.data_in)
                op_queue.add_no_op(
                    description='Not creating IDS',
                    extra_description=f'IDS entry {model.data_in} exists')
                any_exists = True

        return any_exists

    def run_dirs_exist(self, models: Sequence[Run]) -> bool:
        """Check if any of the run dirs exist."""
        any_exists = False

        for model in models:
            if model.dirname.exists():
                op_queue.add_no_op(
                    description='Not creating directory',
                    extra_description=f'Directory {model.dirname} exists',
                )

                any_exists = True

        return any_exists

    def warn_no_create_runs(self):
        """Add warning to op queue if runs will not be created."""
        op_queue.warning(description='Warning',
                         extra_description='Some targets already exist, '
                         'use --force to override')

    @add_to_op_queue('Setting inital condition of', '{data_in}', quiet=True)
    def apply_operations(self, data_in: ImasHandle, run_dir: Path,
                         operations: list[Any]):
        for model in operations:
            apply_model(model, run_dir=run_dir, ids_mapping=data_in)

    @add_to_op_queue('Writing runs', '{self.runs_yaml}', quiet=True)
    def write_runs_file(self, runs: Sequence[Run]) -> None:
        runs = Runs.parse_obj(runs)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            with open(self.runs_yaml, 'w') as f:
                runs.yaml(stream=f)

            # Only if it is a different directory
            if self._is_runs_dir_different_from_config_dir():
                with open(self.runs_dir / 'runs.yaml', 'w') as f:
                    runs.yaml(stream=f)

    @add_to_op_queue('Writing csv', quiet=True)
    def write_runs_csv(self, runs: Sequence[Run]):
        fname = self.data_csv

        prefix = f'{cfg.tag}.' if cfg.tag else ''

        run_map = {
            f'{prefix}{run.shortname}': run.data_out.dict()
            for run in runs
        }
        df = pd.DataFrame.from_dict(run_map, orient='index')
        df.to_csv(fname)

        if self._is_runs_dir_different_from_config_dir():
            df.to_csv(self.runs_dir / fname)

    @add_to_op_queue('Storing duqtools.yaml inside runs_dir', quiet=True)
    def copy_config(self, config):
        if self._is_runs_dir_different_from_config_dir():
            shutil.copyfile(Path.cwd() / config,
                            self.runs_dir / 'duqtools.yaml')

    def _is_runs_dir_different_from_config_dir(self) -> bool:
        """Return True if the runs dir is different from the duqtools config
        dir."""
        return Path.cwd().resolve() != self.runs_dir.resolve()

    def create_run(self, model: Run, *, force: bool = False):
        """Take a run model and create it."""
        op_queue.add(action=model.dirname.mkdir,
                     kwargs={
                         'parents': True,
                         'exist_ok': force
                     },
                     description='Creating run',
                     extra_description=f'{model.dirname}')

        self.source.copy_data_to(model.data_in)

        self.system.copy_from_template(self.template_drc, model.dirname)

        self.apply_operations(model.data_in, model.dirname, model.operations)

        self.system.write_batchfile(model.dirname)

        self.system.update_imas_locations(run=model.dirname,
                                          inp=model.data_in,
                                          out=model.data_out)


def create(*, force, config, **kwargs):
    """Create input for jetto and IDS data structures.

    Parameters
    ----------
    force : bool
        Override protection if data and directories already exist.
    config : Path
        Config file location

    **kwargs
        Unused.
    """
    create_mgr = CreateManager()

    ops_list = create_mgr.generate_ops_list()

    runs = create_mgr.make_run_models(ops_list)

    if not force:

        target_exists = any([
            create_mgr.runs_yaml_exists(),
            create_mgr.data_locations_exist(runs),
            create_mgr.run_dirs_exist(runs),
        ])

        if target_exists:
            create_mgr.warn_no_create_runs()
            return

    for model in runs:
        create_mgr.create_run(model, force=force)

    create_mgr.write_runs_file(runs)
    create_mgr.write_runs_csv(runs)
    create_mgr.copy_config(config)


def recreate(*, runs: Sequence[Path], **kwargs):
    """Create input for jetto and IDS data structures.

    Parameters
    ----------
    runs : Sequence[Path]
        Run names to recreate.
    **kwargs
        Unused.
    """
    create_mgr = CreateManager()

    run_dict = {run.shortname: run for run in Locations().runs}

    run_models = []
    for run in runs:
        if run not in run_dict:
            raise ValueError(f'`{run}` not in `runs.yaml`.')

        model = run_dict[run]
        model.data_in = ImasHandle.parse_obj(model.data_in)
        model.data_out = ImasHandle.parse_obj(model.data_out)

        model.data_in.delete()
        remove_run(model)

        run_models.append(model)

    for model in run_models:
        create_mgr.create_run(model)



import logging
from typing import Optional, Sequence, Union

import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)


def standardize_time(ds: xr.Dataset, *, start: int = 0) -> None:
    """Standardize the time within a dataset.

    Simply subtracts time[0] from all time entries and adds `start`
    Note: this does not interpolate the times between different datasets

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    start : int, optional
        Where to start the returned time series
    """

    ds['time'] = ds['time'] - ds['time'][0] + start


def squash_placeholders(ds: xr.Dataset) -> xr.Dataset:
    """Squash placeholder variables. Data are grouped along the first dimension
    (usually time).

    If the data contains dimensions with a `$`-prefix,
    these are all interpolated to the first array of that type.

    Parameters
    ----------
    ds : xr.Dataset
        xarray Dataset

    Returns
    -------
    ds : xr.Dataset
        xarray Dataset
    """
    prefix = '$'

    dimensions = tuple(str(dim) for dim in ds.dims)

    placeholder_vars = [dim for dim in dimensions if dim.startswith(prefix)]

    for var in placeholder_vars:
        new_dim = var.lstrip(prefix)

        var_index = dimensions.index(var)
        group_dims = dimensions[:var_index]

        groupby = group_dims[0]

        ds = standardize_grid(ds, new_dim=new_dim, old_dim=var, group=groupby)

    return ds


def standardize_grid(ds: xr.Dataset,
                     *,
                     new_dim: str,
                     old_dim: str,
                     group: Optional[str] = None,
                     new_dim_data: Union[np.ndarray, int] = 0) -> xr.Dataset:
    """Standardize the grid within a dataset.

    Perform `split-apply-combine` routine on the data. Split
    by the `group`, standardize the data in `new_dim` using
    `new_dim_data` (interpolate if necessary),
    and combine replacing `old_dim` by `new_dim`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    new_dim : str
        Must be an existing variable with `group` as a dimension.
    old_dim : str
        Must be an existing dimension without coordinates.
    group : str, optional
        Split the data in groups over this dimension.
    new_dim_data : Union[np.ndarray, int], optional
        The data to be used for `new_dim`. If it is an integer,
        use it as an index to grab the data from `new_dim`.

    Returns
    -------
    xr.Dataset
        New dataset with `new_dim` as a coordinate dimension.
    """
    if isinstance(new_dim_data, int):
        new_dim_data = ds.isel(  # type: ignore
            **{group: new_dim_data})[new_dim].data  # type:ignore

    gb = ds.groupby(group)

    interp_kwargs = {new_dim: new_dim_data}

    def standardize(group):
        group = group.swap_dims({old_dim: new_dim})
        group = group.interp(**interp_kwargs)
        return group

    return gb.map(standardize)


def rebase_on_grid(ds: xr.Dataset, *, coord_dim: str,
                   new_coords: np.ndarray) -> xr.Dataset:
    """Rebase (interpolate) the coordinate dimension to the new coordinates.

    Thin wrapper around `xarray.Dataset.interp`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    coord_dim : str
        Name of the grid dimension (i.e. grid variable).
    new_coords : np.ndarray
        The coordinates to interpolate to

    Returns
    -------
    xr.Dataset
        Rebased dataset
    """
    return ds.interp(coords={coord_dim: new_coords},
                     kwargs={'fill_value': 'extrapolate'})


def rebase_on_time(ds: xr.Dataset,
                   *,
                   time_dim='time',
                   new_coords: np.ndarray) -> xr.Dataset:
    """Rebase (interpolate) the time dimension to the new coordinates.

    Thin wrapper around `xarray.Dataset.interp`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    time_dim : str
        Name of the time dimension (i.e. time variable).
    new_coords : np.ndarray
        The coordinates to interpolate to

    Returns
    -------
    xr.Dataset
        Rebased dataset
    """
    if len(ds[time_dim]) < 2:
        # nothing to rebase with only 1 timestep
        return ds
    else:
        return rebase_on_grid(ds, coord_dim=time_dim, new_coords=new_coords)


def standardize_grid_and_time(
    datasets: Sequence[xr.Dataset],
    *,
    grid_var: str = 'rho_tor_norm',
    time_var: str = 'time',
    reference_dataset: int = 0,
) -> tuple[xr.Dataset, ...]:
    """Standardize list of datasets by applying standard rebase operations.

    Applies, in sequence:
    1. `standardize_grid`
    2. `rebase_on_grid`
    3. `rebase_on_time`

    Parameters
    ----------
    datasets : Sequence[xr.Dataset]
        List of source datasets
    grid_var : str, optional
        Name of the grid dimension (i.e. grid variable)
    time_var : str, optional
        Name of the time dimension (i.e. time variable)
    reference_dataset : int, optional
        The dataset with this index will be used as the reference for rebasing.
        The grid and time coordinates of the other datasets will be rebased
        to the reference.

    Returns
    -------
    tuple[xr.Dataset]
        Tuple of output datasets
    """
    reference_grid = datasets[reference_dataset][grid_var].data

    datasets = tuple(
        rebase_on_grid(ds, coord_dim=grid_var, new_coords=reference_grid)
        for ds in datasets)

    reference_time = datasets[reference_dataset][time_var].data

    datasets = tuple(
        rebase_on_time(ds, time_dim=time_var, new_coords=reference_time)
        for ds in datasets)

    return datasets


def rebase_all_coords(
    datasets: Sequence[xr.Dataset],
    reference_dataset: xr.Dataset,
) -> tuple[xr.Dataset, ...]:
    """Rebase all coords, by applying rebase operations.

    Parameters
    ----------
    datasets : Sequence[xr.Dataset]
        datasets
    reference_dataset : xr.Dataset
        reference_dataset

    Returns
    -------
    tuple[xr.Dataset, ...]
    """

    interp_dict = {
        name: dim
        for name, dim in reference_dataset.coords.items() if dim.size > 1
    }

    return tuple(
        ds.interp(coords=interp_dict, kwargs={'fill_value': 'extrapolate'})
        for ds in datasets)



from __future__ import annotations

from typing import Annotated, Literal, Union

from pydantic import Field, validator

from ._basemodel import BaseModel
from ._description_helpers import formatter as f


class JsetField(BaseModel):
    file: Literal['jetto.jset'] = Field('jetto.jset',
                                        description='Name of the file.')
    field: str = Field(description='Field name.')


class NamelistField(BaseModel):
    file: Literal['jetto.in'] = Field('jetto.in',
                                      description='Name of the file.')
    field: str = Field(description='Field name.')
    section: str = Field(description='Section in the config.')

    @validator('section')
    def section_lower(cls, v):
        return v.lower()


JettoField = Annotated[Union[JsetField, NamelistField],
                       Field(discriminator='file')]


class JettoVar(BaseModel):
    """These describe the jetto variables."""
    doc: str = Field(description='Docstring for the variable.')
    name: str = Field(description='Name of the variable.')
    type: Literal['str', 'int', 'float'] = Field(
        description=f('Type of the variable (str, int, float)'))
    keys: list[JettoField] = Field(description=f(
        'Jetto keys to update when this jetto variable is requested'))

    def get_type(self):
        return {
            'str': str,
            'int': int,
            'float': float,
        }[self.type]



import logging
from enum import Enum
from pathlib import Path

import click

from ..config import cfg

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug

cs = click.style

STATUS_SYMBOLS = {
    'no status': cs('_', fg='yellow'),
    'completed': cs('.', fg='green'),
    'failed': cs('f', fg='red'),
    'running': cs('r', fg='yellow'),
    'submitted': cs('s', fg='yellow'),
    'unknown': cs('u', fg='yellow')
}


class JobStatus(str, Enum):
    NOSTATUS = 'no status'
    COMPLETED = 'completed'
    FAILED = 'failed'
    RUNNING = 'running'
    SUBMITTED = 'submitted'
    UNKNOWN = 'unknown'

    @property
    def symbol(self):
        return STATUS_SYMBOLS[self.value]

    @staticmethod
    def symbol_help():
        s = ', '.join(f'{code} : {status}'
                      for status, code in STATUS_SYMBOLS.items())
        return f'Status codes:\n{s}'


class Job:

    def __init__(self, dir: Path):
        self.dir = Path(dir)

    def __repr__(self):
        run = str(self.dir)
        return f'{self.__class__.__name__}({run!r})'

    @staticmethod
    def status_symbol_help():
        """Return help string for status codes."""
        return JobStatus.symbol_help()

    @property
    def status_symbol(self):
        """One letter status symbol."""
        return self.status().symbol

    @property
    def has_submit_script(self) -> bool:
        return self.submit_script.exists()

    @property
    def has_status(self) -> bool:
        return self.status_file.exists()

    @property
    def is_submitted(self) -> bool:
        return (self.dir / 'duqtools.submit.lock').exists()

    def status(self) -> str:
        if not self.has_status:
            return JobStatus.NOSTATUS

        sf = self.status_file
        with open(sf) as f:
            content = f.read()
            if cfg.status.msg_completed in content:
                return JobStatus.COMPLETED
            elif cfg.status.msg_failed in content:
                return JobStatus.FAILED
            elif cfg.status.msg_running in content:
                return JobStatus.RUNNING

        if self.is_submitted:
            return JobStatus.SUBMITTED

        return JobStatus.UNKNOWN

    @property
    def is_completed(self) -> bool:
        return self.status == JobStatus.COMPLETED

    @property
    def is_failed(self) -> bool:
        return self.status == JobStatus.FAILED

    @property
    def is_running(self) -> bool:
        return self.status == JobStatus.RUNNING

    @property
    def in_file(self) -> Path:
        return self.dir / cfg.status.in_file

    @property
    def out_file(self) -> Path:
        return self.dir / cfg.status.out_file

    @property
    def status_file(self) -> Path:
        return self.dir / cfg.status.status_file

    @property
    def submit_script(self) -> Path:
        return self.dir / cfg.submit.submit_script_name

    @property
    def lockfile(self) -> Path:
        return self.dir / 'duqtools.submit.lock'

    def submit(self):
        from ..system import get_system
        debug(f'Put lockfile in place for {self.lockfile}')
        self.lockfile.touch()

        get_system().submit_job(self)

    def start(self):
        click.echo(f'Submitting {self}\033[K')
        self.submit()

        while not self.has_status:
            yield

        while self.is_running:
            yield



# https://github.com/duqtools/duqtools/issues/27

import logging

logger = logging.getLogger(__name__)
imas_mocked = False

try:
    import os
    import xml.sax
    import xml.sax.handler

    import imas
    from imas import imasdef

    IMAS_PREFIX = os.getenv('IMAS_PREFIX', '')
    PATH_IDSDEF = f'{IMAS_PREFIX}/include/IDSDef.xml'

    class Parser(xml.sax.handler.ContentHandler):
        import xml.sax
        import xml.sax.handler

        def __init__(self):
            xml.sax.handler.ContentHandler.__init__(self)
            self.idss = []

        def startElement(self, name: str, attrs):
            if name == 'IDS':
                ids = {}
                for i in attrs.getNames():
                    ids[i] = attrs.getValue(i)
                self.idss.append(ids)

        @classmethod
        def load_idsdef(cls):
            parser = cls()
            xml.sax.parse(PATH_IDSDEF, parser)
            return parser

except (ModuleNotFoundError, ImportError):
    from unittest.mock import MagicMock as Mock
    imas_mocked = True
    ids = Mock()
    ids.open_env = lambda *_, **__: [1]

    entry = Mock()
    entry.open = lambda *_, **__: (0, Mock())

    imas = Mock()
    imas.names = ['imas_3_34_0_ual_4_9_3']
    imas.ids = lambda *_, **__: ids
    imas.DBEntry = lambda *_, **__: entry

    imasdef = Mock()

    Parser = Mock()  # type: ignore

    logger.warning('Could not import IMAS:', exc_info=True)



import logging
import sys
from pathlib import Path

if sys.version_info < (3, 10):
    from importlib_resources import files
else:
    from importlib.resources import files

logger = logging.getLogger(__name__)


def dash(**kwargs):
    """Start streamlit dashboard."""
    try:
        from streamlit.web import cli as stcli
    except ImportError:
        # versions <= (1.11)
        from streamlit import cli as stcli

    dashboard_path = files('duqtools.data') / 'dash' / 'dash.py'

    workdir = Path('.').resolve()

    sys.argv = ['streamlit', 'run', str(dashboard_path), '--', str(workdir)]

    logger.debug('Streamlit arguments %s', sys.argv)

    sys.exit(stcli.main())



from typing import Sequence

import xarray as xr

from ..config import var_lookup
from ..operations import add_to_op_queue
from ..schema import IDSVariableModel
from ..utils import groupby
from ._handle import ImasHandle
from ._rebase import rebase_all_coords, squash_placeholders


@add_to_op_queue('Merging to', '{target}')
def merge_data(
    handles: Sequence[ImasHandle],
    target: ImasHandle,
    variables: list[IDSVariableModel],
    callback=None,
):
    """merge_data merges the data from the handles to the target, only merges
    over the listed variables, coordination variables are never overwritten,
    and data is rebased according to the target coordination variable.

    Parameters
    ----------
    handles : Sequence[ImasHandle]
        handles
    target : ImasHandle
        target
    variables : Sequence[IDSVariableModel]
        variables
    """
    # Add dimensions to variables
    variable_dict = {}

    for variable in variables:
        variable_dict[variable.name] = variable

        for dim in variable.dims:
            try:
                dim_var = var_lookup[dim]
                variable_dict.setdefault(dim_var.name, dim_var)
            except KeyError:
                # skip dimensions without coordinates like `ion`
                pass

    variables = tuple(variable_dict.values())

    # Get all known variables per ids
    grouped_ids_vars = groupby(variables, keyfunc=lambda var: var.ids)

    i_tot = len(grouped_ids_vars)

    for i, (ids_name, ids_vars) in enumerate(grouped_ids_vars.items()):
        if callback:
            callback(i / i_tot)

        # Get all data, and rebase it
        target_ids = target.get(ids_name)  # type: ignore
        target_data = target_ids.to_xarray(variables=ids_vars,
                                           empty_var_ok=True)
        target_data = squash_placeholders(target_data)

        ids_data = [
            handle.get_variables(ids_vars, empty_var_ok=True)  # type: ignore
            for handle in handles
        ]

        ids_data = rebase_all_coords(ids_data, target_data)

        ids_data = xr.concat(ids_data, 'handle')

        # Now we have to get the stddeviations
        mean_data = ids_data.mean(dim='handle')
        std_data = ids_data.std(dim='handle', skipna=True)

        # Then, write it back to target
        for name in ids_data.data_vars.keys():
            target_ids.write_array_in_parts(variable_dict[name].path,
                                            mean_data[name])
            target_ids.write_array_in_parts(
                variable_dict[name].path + '_error_upper', std_data[name])

        target_ids.sync(target)



import functools
import logging
from datetime import datetime
from pathlib import Path
from sys import exit, stderr, stdout
from typing import Callable

import click
from pydantic import ValidationError

from ._click_opt_groups import GroupCmd, GroupOpt
from ._logging_utils import TermEscapeCodeFormatter, duqlog_screen
from .config import cfg
from .operations import op_queue, op_queue_context

logger = logging.getLogger(__name__)

try:
    import coverage
    coverage.process_startup()
except ImportError:
    pass


def config_option(f):
    return click.option('-c',
                        '--config',
                        default='duqtools.yaml',
                        help='Path to config.',
                        cls=GroupOpt,
                        group='Common options')(f)


def quiet_option(f):
    return click.option('-q',
                        '--quiet',
                        is_flag=True,
                        default=False,
                        help='Don\'t output anything to the screen'
                        ' (except mandatory prompts).',
                        cls=GroupOpt,
                        group='Common options')(f)


def debug_option(f):
    """Must be added together with `logfile_option`."""
    return click.option('--debug',
                        is_flag=True,
                        help='Enable debug print statements.',
                        cls=GroupOpt,
                        group='Common options')(f)


def dry_run_option(f):
    return click.option('--dry-run',
                        is_flag=True,
                        help='Execute without any side-effects.',
                        cls=GroupOpt,
                        group='Common options')(f)


def yes_option(f):
    return click.option('--yes',
                        is_flag=True,
                        help='Answer yes to questions automatically.',
                        cls=GroupOpt,
                        group='Common options')(f)


def logfile_option(f):
    """Must be added together with `debug_option`."""
    return click.option('--logfile',
                        '-l',
                        is_flag=False,
                        default='duqtools.log',
                        help='where to send the logfile,'
                        ' the special values stderr/stdout'
                        ' will send it there respectively.',
                        cls=GroupOpt,
                        group='Common options')(f)


logging_options = (logfile_option, debug_option)
all_options = (*logging_options, config_option, quiet_option, dry_run_option,
               yes_option)


def handles_option(f):
    return click.option('-h',
                        '--handle',
                        'handles',
                        type=str,
                        help='IMAS data handles.',
                        multiple=True)(f)


def variables_option(f):
    return click.option('-v',
                        '--variable',
                        'var_names',
                        type=str,
                        help='Name of the variables.',
                        multiple=True)(f)


def datafile_option(f):
    return click.option('-i',
                        '--input',
                        'input_files',
                        type=str,
                        help='Input file, i.e. `data.csv` or `runs.yaml`',
                        multiple=True)(f)


class OptionParser:

    def wrap(self, func: Callable):
        """With this wrapper it becomes possible to parse common options in
        user defined order."""

        def callback(**kwargs):
            for parse in (
                    self.parse_config,
                    self.parse_logfile,
                    self.parse_yes,
                    self.parse_dry_run,
                    self.parse_quiet,
            ):
                try:
                    parse(**kwargs)
                except TypeError:
                    # Skip when option has not been set
                    pass
            func(**kwargs)

        return callback

    def parse_logfile(self, *, logfile, debug, **kwargs):
        level = logging.DEBUG if debug else logging.INFO

        streams = {'stdout': stdout, 'stderr': stderr}
        logging.getLogger().handlers = []

        if logfile in streams.keys():
            logging.basicConfig(stream=streams[logfile], level=level)
        else:
            fhandler = logging.FileHandler(logfile)
            fhandler.setLevel(level)

            # Remove fancies from logfiles
            escaped_format = TermEscapeCodeFormatter(logging.BASIC_FORMAT)
            fhandler.setFormatter(escaped_format)
            logging.getLogger().addHandler(fhandler)

        logger.info('')
        logger.info(
            'Duqtools starting at '
            f'{datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S %z")}')
        logger.info('------------------------------------------------')
        logger.info('')

    def parse_quiet(self, *, quiet, **kwargs):
        if quiet:
            duqlog_screen.handlers = []  # remove output methods
            cfg.quiet = True

    def parse_dry_run(self, *, dry_run, **kwargs):
        op_queue.dry_run = dry_run

    def parse_config(self, *, config, **kwargs):
        try:
            cfg.parse_file(config)
        except ValidationError as e:
            exit(e)

    def parse_yes(self, *, yes, **kwargs):
        op_queue.yes = yes


def common_options(*options):
    """common_options.

    IMPORTANT: This must be the last click option specified before
    the execution of the actual function, otherwise options will be
    missing
    """

    def decorator(func):
        wrapper = OptionParser().wrap(func)

        for option in options:
            wrapper = option(wrapper)

        functools.update_wrapper(wrapper, func)

        return wrapper

    return decorator


def cli_entry(**kwargs):
    from duqtools import fix_dependencies
    fix_dependencies()
    cli()


@click.group()
def cli(**kwargs):
    """For more information, check out the documentation:

    https://duqtools.readthedocs.io
    """
    pass


@cli.command('init', cls=GroupCmd)
@click.option('-o',
              '--out',
              'out_file',
              help='Path to write config to (default=duqtools.yaml).',
              default='duqtools.yaml')
@click.option('--force', is_flag=True, help='Overwrite existing config.')
@common_options(*logging_options, quiet_option, dry_run_option, yes_option)
def cli_init(**kwargs):
    """Create a default config file."""
    from .init import init
    with op_queue_context():
        try:
            init(**kwargs)
        except RuntimeError as e:
            exit(e)


@cli.command('create', cls=GroupCmd)
@click.option('--force',
              is_flag=True,
              help='Overwrite existing run directories and IDS data.')
@common_options(*all_options)
def cli_create(**kwargs):
    """Create the UQ run files."""
    from .create import create
    with op_queue_context():
        create(**kwargs)


@cli.command('recreate', cls=GroupCmd)
@click.argument('runs', type=Path, nargs=-1)
@common_options(*all_options)
def cli_recreate(**kwargs):
    """Read `runs.yaml` and re-create the given runs.

    \b
    Example:

    - `duqtools recreate run_0003 run_0004 --force`
    """
    from .create import recreate
    with op_queue_context():
        recreate(**kwargs)


@cli.command('submit', cls=GroupCmd)
@click.option('--force',
              is_flag=True,
              help='Re-submit running or completed jobs.')
@click.option(
    '--schedule',
    is_flag=True,
    help='Schedule and submit jobs automatically. `max_jobs` must be defined.')
@click.option('-j',
              '--max_jobs',
              type=int,
              help='Maximum number of jobs to submit.')
@click.option('-a', '--array', is_flag=True, help='Submit jobs as array.')
@click.option('-r',
              '--resubmit',
              multiple=True,
              default=tuple(),
              type=Path,
              help='Case to re-submit, can be specified multiple times')
@common_options(*all_options)
def cli_submit(**kwargs):
    """Submit the UQ runs.

    This subcommand will read `runs.yaml`, and start all runs which are not yet
    running. By default, It will not re-submit running or finished jobs.

    There is a scheduler that will continue to submit jobs until the specified
    maximum number of jobs is running. Once a job has completed, a new job will
    be submitted from the queue to fill the spot.
    """
    from .submit import submit
    with op_queue_context():
        submit(**kwargs)


@cli.command('status', cls=GroupCmd)
@click.option('--detailed', is_flag=True, help='Detailed info on progress')
@click.option('--progress', is_flag=True, help='Fancy progress bar')
@common_options(*all_options)
def cli_status(**kwargs):
    """Print the status of the UQ runs."""
    from .status import status
    status(**kwargs)


@cli.command('plot')
@handles_option
@variables_option
@click.option('-o',
              '--format',
              'extensions',
              type=str,
              help='Output format (json, html, png, svg, pdf), default: html.',
              default=('html', ),
              multiple=True)
@click.option('-e',
              '--errorbars',
              is_flag=True,
              help='Plot the errorbars (if present)')
@datafile_option
def cli_plot(**kwargs):
    """Plot some IDS data.

    \b
    Examples:

    - `duqtools plot -v t_i_ave -v zeff -i data.csv`
    - `duqtools plot -v t_i_ave -v zeff -h user/jet/91234/5`
    - `duqtools plot -v zeff -h db/91234/5 -h db/91234/6 -h db/91234/7`
    - `duqtools plot -v zeff -h db/91234/5 -i data.csv`
    - `duqtools plot -v zeff -h db/91234/5 -o json`
    """
    from .plot import plot
    plot(**kwargs)


@cli.command('clean', cls=GroupCmd)
@click.option('--out', is_flag=True, help='Remove output data.')
@click.option('--force',
              is_flag=True,
              help='Overwrite backup file if necessary.')
@common_options(*all_options)
def cli_clean(**kwargs):
    """Delete generated IDS data and the run dir."""
    from .cleanup import cleanup
    with op_queue_context():
        cleanup(**kwargs)


@cli.command('go', cls=GroupCmd)
@click.option('--force', is_flag=True, help='Overwrite files when necessary.')
@common_options(*all_options)
def cli_go(**kwargs):
    """Run create, submit, status, dash in succession.

    Useful for existing tested and working pipelines.
    """
    from .create import create
    from .dash import dash
    from .status import status
    from .submit import submit
    with op_queue_context():
        create(**kwargs)
    with op_queue_context():
        submit(max_jobs=None,
               array=True,
               schedule=None,
               resubmit=tuple(),
               **kwargs)

    skwargs = kwargs.copy()
    skwargs['detailed'] = True
    skwargs['progress'] = False
    status(**skwargs)

    dash(**kwargs)


@cli.command('yolo')
@click.pass_context
def cli_yolo(ctx, **kwargs):
    """Live on the edge, run `duqtools go --force --yes --quiet`."""
    ctx.invoke(cli_go, force=True, quiet=True, yes=True)


@cli.command('dash', cls=GroupCmd)
@common_options(*logging_options, quiet_option, dry_run_option, yes_option)
def cli_dash(**kwargs):
    """Open dashboard for evaluating IDS data."""
    from .dash import dash
    dash(**kwargs)


@cli.command('merge', cls=GroupCmd)
@handles_option
@variables_option
@click.option('-t',
              '--template',
              required=True,
              type=str,
              help='IMAS location to use as the template for the target')
@click.option('-o',
              '--out',
              'target',
              required=True,
              type=str,
              help='IMAS location to store the result in')
@click.option('--all',
              'merge_all',
              is_flag=True,
              help='Try to merge all known variables.')
@datafile_option
@click.option('--force',
              is_flag=True,
              help='Overwrite existing output dataset.')
@common_options(*all_options)
def cli_merge(**kwargs):
    """Merge data sets with error propagation.

    Example Merging two IDSes. Run number `8000` is the template.
    Only The `t_e` variable is merged.

    The resulting IDS is saved to your own test database with
    shot number `36982` and run number `9999`.

    Example:

    \b
    ```
    duqtools merge \\
        -t g2jcitri/aug/36982/8000 \\
        -o test/36982/9999 \\
        -h g2jcitri/aug/36982/8001 \\
        -h g2jcitri/aug/36982/8000 \\
        -v t_e
    ```
    \f
    Note:

    The `-t`, `-o` and `-h` options expect an IMAS path formatted as
    `user/db/shot/number` or `db/shot/number`.
    """
    from .merge import merge
    with op_queue_context():
        merge(**kwargs)


@cli.command('list-variables')
@config_option
def cli_list_variables(**kwargs):
    """List available variables.

    Picks up variables from `duqtools.yaml` if it exists in the local
    directory.
    """
    from .list_variables import list_variables
    list_variables(**kwargs)


@cli.command('version')
def cli_version(**kwargs):
    """Print the version and exit."""
    import git

    from duqtools import __version__

    string = f'duqtools {__version__}'

    try:
        repo = git.Repo(Path(__file__), search_parent_directories=True)
        sha = repo.head.object.hexsha
    except OSError:
        pass
    else:
        string += f' (rev: {sha})'

    click.echo(string)


if __name__ == '__main__':
    cli()



import logging

from ._handle import ImasHandle
from ._imas import imas_mocked
from ._mapping import IDSMapping
from ._merge import merge_data
from ._rebase import (
    rebase_all_coords,
    rebase_on_grid,
    rebase_on_time,
    squash_placeholders,
    standardize_grid,
    standardize_grid_and_time,
    standardize_time,
)

logger = logging.getLogger(__name__)

__all__ = [
    'IDSMapping',
    'ImasHandle',
    'merge_data',
    'rebase_on_grid',
    'rebase_on_time',
    'rebase_all_coords',
    'standardize_grid_and_time',
    'standardize_grid',
    'standardize_time',
    'squash_placeholders',
    'imas_mocked',
]



"""Function to create llcmd file."""
from __future__ import annotations

import stat
from pathlib import Path

import jetto_tools

from ..models import Locations


def write_batchfile(run_dir: Path,
                    jset: jetto_tools.jset.JSET,
                    tag: str | None = None):
    """Write batchfile (`.llcmd`) to start jetto.

    Parameters
    ----------
    run_dir : Path
        Directory to place batch file into.
    jset: JSET
        Jetto settings object (from jetto-pythontools)
    tag : str | None
        Tag the job, defaults to 'jetto'.
    """
    if not tag:
        tag = 'jetto'

    llcmd_path = run_dir / '.llcmd'

    rjettov_path = run_dir / 'rjettov'
    rel_path = run_dir.relative_to(Locations().jruns_path.resolve())

    build_name = jset['JobProcessingPanel.name']
    build_user_name = jset['JobProcessingPanel.userid']
    machine_number = jset['JobProcessingPanel.machineNumber']
    num_proc = jset['JobProcessingPanel.numProcessors']
    wall_time = jset['JobProcessingPanel.wallTime']

    with open(llcmd_path, 'w') as f:
        f.write(f"""#!/bin/sh
#SBATCH -J duqtools.{tag}.{run_dir.name}
#SBATCH -i /dev/null
#SBATCH -o {run_dir / 'll.out'}
#SBATCH -e {run_dir / 'll.err'}
#SBATCH -p {machine_number}

#SBATCH -N 1
#SBATCH -n {num_proc}
#SBATCH -t {wall_time}:00:00

cd {run_dir}
{rjettov_path} -S -I -p -xmpi -x64 {rel_path} \
{build_name} {build_user_name}""")

    llcmd_path.chmod(llcmd_path.stat().st_mode | stat.S_IXUSR)



from __future__ import annotations

from pathlib import Path
from typing import Union

from pydantic import Field, root_validator

from ._basemodel import BaseModel
from ._dimensions import IDSOperation, JettoOperation
from ._imas import ImasBaseModel


class Run(BaseModel):
    dirname: Path = Field(description='Directory of run')
    shortname: Path = Field(description='Short name (`dirname.name`)')
    data_in: ImasBaseModel
    data_out: ImasBaseModel
    operations: list[Union[IDSOperation, JettoOperation,
                           list[Union[IDSOperation, JettoOperation]]]]

    @root_validator()
    def shortname_compat(cls, root):
        # Compatibility with old runs.yaml
        if not root['shortname']:
            root['shortname'] = root['dirname'].name
        return root


class Runs(BaseModel):
    __root__: list[Run] = []

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index: int):
        return self.__root__[index]



"""This module contains tools for interfacing with jetto runs."""

from ._jettovar_to_json import jettovar_to_json
from ._llcmd import write_batchfile
from ._system import BaseJettoSystem, JettoSystem, JettoSystemV210921, JettoSystemV220922

__all__ = [
    'BaseJettoSystem',
    'JettoSystem',
    'JettoSystemV210921',
    'JettoSystemV220922',
    'jettovar_to_json',
    'write_batchfile',
]



from pathlib import Path

from .config import cfg
from .ids import ImasHandle
from .jetto import JettoSystemV210921, JettoSystemV220922
from .models import AbstractSystem, Job


class DummySystem(AbstractSystem):
    """This is a dummy system that implements the basic interfaces.

    It exists for testing purposes in absence of actual modelling
    software.
    """

    @staticmethod
    def get_runs_dir() -> Path:
        return Path()

    @staticmethod
    def write_batchfile(run_dir: Path):
        pass

    @staticmethod
    def submit_job(job: Job):
        pass

    @staticmethod
    def copy_from_template(source_drc: Path, target_drc: Path):
        pass

    @staticmethod
    def imas_from_path(template_drc: Path):
        return ImasHandle(db='', shot='-1', run='-1')

    @staticmethod
    def update_imas_locations(run: Path, inp, out):
        pass


def get_system():
    """get_system.

    Get the system to do operations with TODO make it a variable, not a
    function
    """
    if (cfg.system in ['jetto', 'jetto-v220922']):
        return JettoSystemV220922
    elif (cfg.system in ['jetto-v210921']):
        return JettoSystemV210921
    elif (cfg.system == 'dummy'):
        return DummySystem
    else:
        raise NotImplementedError(f'system {cfg.system} is not implemented')



from pathlib import Path

from ..config import cfg
from ..merge import merge as duqtools_merge


def merge(**kwargs):
    cwd = Path.cwd()

    config_files = cwd.glob('**/duqtools.yaml')

    for config_file in config_files:
        cfg.parse_file(config_file)

        duqtools_merge()



from __future__ import annotations

from typing import Literal, Union

from pydantic import Field, validator

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._ranges import ARange, LinSpace
from ._variable import IDSVariableModel, JettoVariableModel


class OperatorMixin(BaseModel):
    operator: Literal['add', 'multiply', 'divide', 'power', 'subtract',
                      'floor_divide', 'mod', 'copyto',
                      'remainder'] = Field('multiply',
                                           description=f("""
        Which operator to apply to the data in combination with any of the
        given values below. This can be any of the basic numpy arithmetic
        operations. Available choices: `add`, `multiply`, `divide`, `power`,
        `subtract`, `floor_divide`, `mod`, `none` and `remainder`. These directly map
        to the equivalent numpy functions, i.e. `add` -> `np.add`.
        """))
    scale_to_error: bool = Field(False,
                                 description=f("""
        If True, multiply value(s) by the error (sigma).

        With asymmetric errors (i.e. both lower/upper error nodes are available),
        scale to the lower error node for values < 0, and to the upper error node
        for values > 0.
        """))

    _upper_suffix: str = '_error_upper'
    _lower_suffix: str = '_error_lower'


class DimMixin(BaseModel):
    values: Union[list[float], ARange, LinSpace] = Field(description=f("""
            Values to use with operator on field to create sampling
            space."""))

    @validator('values')
    def convert_to_list(cls, v):
        if not isinstance(v, list):
            v = v.values
        return v


class OperationDim(OperatorMixin, DimMixin, BaseModel):
    variable: str = Field(description=f("""
    PlaceHolder for the actual Operator
    """))

    def expand(self, *args, **kwargs):
        from duqtools.config import var_lookup
        variable = var_lookup[self.variable]

        if isinstance(variable, JettoVariableModel):
            return JettoOperationDim.expand(self,
                                            *args,
                                            variable=variable,
                                            **kwargs)
        elif isinstance(variable, IDSVariableModel):
            return IDSOperationDim.expand(self,
                                          *args,
                                          variable=variable,
                                          **kwargs)
        else:
            raise NotImplementedError(
                f'{self.variable} expand not implemented')


class CoupledDim(BaseModel):
    __root__: list[OperationDim]

    @validator('__root__')
    def check_dimensions_match(cls, dims):
        if len(dims) > 0:
            refdim = len(dims[0].values)
            for dim in dims[1:]:
                if not len(dim.values) == refdim:
                    raise ValueError('dimensions do not match in coupled dim')
        return dims

    def expand(self, *args, **kwargs):
        expanded = [operation.expand() for operation in self.__root__]
        return [entry for entry in zip(*expanded)]  # Transpose


class IDSPathMixin(BaseModel):
    variable: IDSVariableModel = Field(description=f("""
            IDS variable for the data to modify.
            The time slice can be denoted with '*', this will match all
            time slices in the IDS. Alternatively, you can specify the time
            slice directly, i.e. `profiles_1d/0/t_i_ave` to only
            match and update the 0-th time slice.
            """))


class IDSOperation(IDSPathMixin, OperatorMixin, BaseModel):
    """Apply arithmetic operation to IDS."""

    value: float = Field(description=f("""
        Values to use with operator on field to create sampling
        space."""))


class IDSOperationDim(IDSPathMixin, OperatorMixin, DimMixin, BaseModel):
    """Apply set of arithmetic operations to IDS.

    Takes the IDS data and subtracts, adds, multiplies, etc with each
    the given values.
    """

    def expand(self, *args, variable, **kwargs) -> tuple[IDSOperation, ...]:
        """Expand list of values into operations with its components."""
        return tuple(
            IDSOperation(variable=variable,
                         operator=self.operator,
                         value=value,
                         scale_to_error=self.scale_to_error)
            for value in self.values)


class JettoPathMixin(BaseModel):
    variable: JettoVariableModel


class JettoOperation(JettoPathMixin, OperatorMixin, BaseModel):
    value: float = Field(description=f("""
        Values to use with operator on field to create sampling
        space."""))


class JettoOperationDim(JettoPathMixin, OperatorMixin, DimMixin, BaseModel):

    def expand(self, *args, variable, **kwargs) -> tuple[JettoOperation, ...]:
        """Expand list of values into operations with its components."""
        return tuple(
            JettoOperation(variable=variable,
                           operator=self.operator,
                           value=value,
                           scale_to_error=self.scale_to_error)
            for value in self.values)



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING

import yaml
from jinja2 import Environment, FileSystemLoader

from duqtools.config import var_lookup

from ..config import Config
from ..operations import op_queue
from ..utils import no_op, read_imas_handles_from_file

if TYPE_CHECKING:
    import jinja2

    from duqtools.api import IDSMapping, ImasHandle

logger = logging.getLogger(__name__)


class SetupError(Exception):
    ...


def get_template(filename: str) -> jinja2.Template:
    """Load filename as a jinja2 template."""
    path = Path(filename)
    drc = Path(path).parent
    file_loader = FileSystemLoader(str(drc))
    environment = Environment(loader=file_loader, autoescape=True)
    return environment.get_template(path.name)


def _get_key(filename: str, *, key: str):
    """Grab key from unrendered config file."""
    with open(filename, 'rb') as f:
        for line in f:
            try:
                item = yaml.safe_load(line)
                return item[key]
            except (yaml.YAMLError, TypeError, KeyError):
                continue


def _generate_run_dir(drc: Path, cfg: str, force: bool):
    drc.mkdir(exist_ok=force, parents=True)

    with open(drc / 'duqtools.yaml', 'w') as f:
        f.write(cfg)


class ExtrasV210921:
    """Track run number to avoid overwriting existing data in sequential
    runs."""
    MAX_RUN = 9999

    def __init__(self, template_file: str):
        self.n_samples = self._get_n_samples(template_file)
        self.run_numbers: dict[tuple[str, int],
                               int] = defaultdict(lambda: 1000)

    @staticmethod
    def _get_n_samples(template_file: str) -> int:
        """Grab number of samples generated by this config, from
        `create.sampler.n_samples`."""
        n_samples = _get_key(template_file, key='n_samples')

        if not n_samples:
            raise ValueError('`create.sampler.n_samples` must be defined for '
                             'jintrac `v210921` config')

        return n_samples

    def add_system_attrs(self, handle: ImasHandle, run: SimpleNamespace):
        """Add system specific attributes to run namespace."""
        data_in_start = self.run_numbers[handle.db, handle.shot]
        data_out_start = data_in_start + self.n_samples

        self.run_numbers[handle.db,
                         handle.shot] = data_out_start + self.n_samples

        if self.run_numbers[handle.db, handle.shot] > self.MAX_RUN:
            raise ValueError(
                f'Cannot write data with run number > {self.MAX_RUN}')

        run.data_in_start = data_in_start
        run.data_out_start = data_out_start


class Variables:
    lookup = var_lookup.filter_type('IDS2jetto-variable')

    def __init__(self, *, handle: ImasHandle):
        self.handle = handle
        self._ids_cache: dict[str, IDSMapping] = {}

    def _get_ids(self, ids: str):
        """Cache ids lookups to avoid repeated data reads."""
        if ids in self._ids_cache:
            mapping = self._ids_cache[ids]
        else:
            mapping = self.handle.get(ids)
            self._ids_cache[ids] = mapping

        return mapping

    def __getattr__(self, key: str):
        try:
            spec = self.lookup[f'ids-{key}']
        except KeyError as exc:
            msg = f'Cannot find {key!r} in your variable listing (i.e. `variables.yaml`).'
            raise AttributeError(msg) from exc

        value = spec.default

        for item in spec.paths:
            mapping = self._get_ids(item.ids)
            try:
                trial = mapping[item.path]
            except KeyError:
                continue

            if all(condition(trial) for condition in spec.accept_if):
                value = trial
                break

        if value is None:
            raise AttributeError(
                f'No value matches specifications given by: {spec}')

        return value


def setup(*, template_file, input_file, force, **kwargs):
    """Setup large scale validation runs for template."""
    cwd = Path.cwd()

    if not input_file:
        raise OSError('Input file not defined.')

    handles = read_imas_handles_from_file(input_file)

    template = get_template(template_file)

    if _get_key(template_file, key='system') == 'jetto-v210921':
        add_system_attrs = ExtrasV210921(template_file).add_system_attrs
    else:
        add_system_attrs = no_op  # default to no-op

    for name, handle in handles.items():
        run = SimpleNamespace(name=name)

        add_system_attrs(run)

        variables = Variables(handle=handle)

        cfg = template.render(run=run, variables=variables, handle=handle)

        Config.parse_raw(cfg)  # make sure config is valid

        out_drc = cwd / name

        if out_drc.exists() and not force:
            op_queue.add_no_op(description='Directory exists',
                               extra_description=name)
            op_queue.warning(description='Warning',
                             extra_description='Some targets already exist, '
                             'use --force to override')
        else:
            op_queue.add(
                action=_generate_run_dir,
                kwargs={
                    'drc': out_drc,
                    'cfg': cfg,
                    'force': force
                },
                description='Setup run',
                extra_description=name,
            )



from pathlib import Path
from typing import Literal, Optional, Union

from pydantic import DirectoryPath, Field, validator

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._dimensions import CoupledDim, OperationDim
from ._imas import ImasBaseModel
from .data_location import DataLocation
from .matrix_samplers import CartesianProduct, HaltonSampler, LHSSampler, SobolSampler
from .variables import VariableConfigModel


class CreateConfigModel(BaseModel):
    """The options of the `create` subcommand are stored in the `create` key in
    the config."""
    dimensions: list[Union[CoupledDim, OperationDim]] = Field(description=f("""
        The `dimensions` specifies the dimensions of the matrix to sample
        from. Each dimension is a compound set of operations to apply.
        From this, a matrix all possible combinations is generated.
        Essentially, it generates the
        [Cartesian product](en.wikipedia.org/wiki/Cartesian_product)
        of all operations. By specifying a different `sampler`, a subset of
        this hypercube can be efficiently sampled.
        """))

    sampler: Union[LHSSampler, HaltonSampler, SobolSampler,
                   CartesianProduct] = Field(default=CartesianProduct(),
                                             discriminator='method',
                                             description=f("""
        For efficient UQ, it may not be necessary to sample the entire matrix
        or hypercube. By default, the cartesian product is taken
        (`method: cartesian-product`). For more efficient sampling of the space,
        the following `method` choices are available:
        [`latin-hypercube`](en.wikipedia.org/wiki/Latin_hypercube_sampling),
        [`sobol`](en.wikipedia.org/wiki/Sobol_sequence),
        [`halton`](en.wikipedia.org/wiki/Halton_sequence).
        Where `n_samples` gives the number of samples to extract.
        """))

    template: DirectoryPath = Field(description=f("""
        Template directory to modify. Duqtools copies and updates the settings
        required for the specified system from this directory. This can be a
        directory with a finished run, or one just stored by JAMS (but not yet
        started). By default, duqtools extracts the input IMAS database entry
        from the settings file (e.g. jetto.in) to find the data to modify for
        the UQ runs.
        """))

    template_data: Optional[ImasBaseModel] = Field(description=f("""
        Specify the location of the template data to modify. This overrides the
        location of the data specified in settings file in the template
        directory.
        """))

    data: Optional[DataLocation] = Field(description=f("""
        Required for `system: jetto-v210921`, irrelevant for other systems.

        Where to store the in/output IDS data.
        The data key specifies the machine or imas
        database name where to store the data (`imasdb`). duqtools will write the input
        data files for UQ start with the run number given by `run_in_start_at`.
        The data generated by the UQ runs (e.g. from jetto) will be stored
        starting by the run number given by `run_out_start_at`.

        """))

    jruns: Optional[DirectoryPath] = Field(description=f(
        """`jruns` defines the the root directory where all simulations are
    run for the jetto system. Because the jettos system works with relative
    directories from some root directory.

    If this variable is not specified, duqtools will look for the `$JRUNS` environment
    variable, and set it to that. If that fails, `jruns` is set to the current directory `./`

    In this way, duqtools can ensure that the current work directory is
    a subdirectory of the given root directory. All subdirectories are
    calculated as relative to the root directory.

    For example, for `rjettov`, the root directory must be set to
    `/pfs/work/$USER/jetto/runs/`. Any UQ runs must therefore be
    a subdirectory.

    """))

    runs_dir: Optional[Path] = Field(description=f(
        """Relative location from the workspace, which specifies the folder where to
    store all the created runs.

    This defaults to `workspace/duqtools_experiment_x`
    where `x` is a not yet existing integer."""))


class SubmitConfigModel(BaseModel):
    """The options of the `submit` subcommand are stored under the `submit` key
    in the config.

    The config describes the commands to start the UQ runs.
    """

    submit_script_name: str = Field(
        '.llcmd', description='Name of the submission script.')
    submit_command: str = Field('sbatch',
                                description='Submission command for slurm.')
    docker_image: str = Field('jintrac-imas',
                              description='Docker image used for submission')
    submit_system: Literal['prominence', 'slurm', 'docker'] = Field(
        'slurm',
        description='System to submit jobs to '
        '[slurm (default), prominence, docker]')


class StatusConfigModel(BaseModel):
    """The options of the `status` subcommand are stored under the `status` key
    in the config.

    These only need to be changed if the modeling software changes.
    """

    status_file: str = Field('jetto.status',
                             description='Name of the status file.')
    in_file: str = Field('jetto.in',
                         description=f("""
            Name of the modelling input file, will be used to check
            if the subprocess has started.
            """))

    out_file: str = Field('jetto.out',
                          description=f("""
            Name of the modelling output file, will be used to
            check if the software is running.
            """))

    msg_completed: str = Field('Status : Completed successfully',
                               description=f("""
            Parse `status_file` for this message to check for
            completion.
            """))

    msg_failed: str = Field('Status : Failed',
                            description=f("""
            Parse `status_file` for this message to check for
            failures.
            """))

    msg_running: str = Field('Status : Running',
                             description=f("""
            Parse `status_file` for this message to check for
            running status.
            """))


class ConfigModel(BaseModel):
    """The options for the CLI are defined by this model."""
    tag: str = Field(
        '',
        description=
        'Create a tag for the runs to identify them in slurm or `data.csv`')

    submit: SubmitConfigModel = Field(
        SubmitConfigModel(),
        description='Configuration for the submit subcommand')

    create: Optional[CreateConfigModel] = Field(
        description='Configuration for the create subcommand')

    status: StatusConfigModel = Field(
        StatusConfigModel(),
        description='Configuration for the status subcommand')

    workspace: Optional[str] = Field(description='Old field, currently unused')

    extra_variables: Optional[VariableConfigModel] = Field(
        description='Specify extra variables for this run.')

    system: Literal['jetto', 'dummy', 'jetto-v210921',
                    'jetto-v220922'] = Field(
                        'jetto', description='backend system to use')

    quiet: bool = Field(
        False,
        description='dont output to stdout, except for mandatory prompts')

    @validator('workspace', pre=True)
    def workspace_deprecated(cls, v):
        if not v:
            from warnings import warn
            warn(f("""workspace key is Deprecated and unused, use create->jruns
            and create->runs_dir to control where runs end up"""),
                 DeprecationWarning,
                 stacklevel=2)
        return None



from pydantic import BaseModel as PydanticBaseModel
from pydantic import Extra
from pydantic_yaml import YamlModelMixin


class BaseModel(YamlModelMixin, PydanticBaseModel):
    """Base model."""

    class Config:
        extra = Extra.forbid



"""Module with click subclasses to group options in `--help`output.

Example:

    python _click_opt_groups.py --help

Implementation based on:
https://github.com/pallets/click/issues/373#issuecomment-515293746
"""

from collections import defaultdict

import click


class GroupOpt(click.Option):

    def __init__(self, *args, **kwargs):
        self.group = kwargs.pop('group', None)
        super().__init__(*args, **kwargs)


class GroupCmd(click.Command):

    def format_options(self, ctx, formatter):
        """Writes all the options into the formatter if they exist.

        Implementation based on:
            https://github.com/pallets/click/issues/373#issuecomment-515293746
        """
        sections = defaultdict(list)

        for param in self.get_params(ctx):
            option = param.get_help_record(ctx)
            if option:
                try:
                    title = str(param.group)
                except AttributeError:
                    title = 'Options'

                sections[title].append(option)

        for title, options in sections.items():
            with formatter.section(title):
                formatter.write_dl(options)


if __name__ == '__main__':

    @click.group()
    def cli():
        pass

    @cli.command(cls=GroupCmd)
    @click.option('-a')
    @click.option('-b', cls=GroupOpt, group='Group 1')
    @click.option('-c', cls=GroupOpt, group='Group 1')
    @click.option('-d', cls=GroupOpt, group='Group 2')
    @click.option('-e', cls=GroupOpt, group='Group 2')
    def main(**kwargs):
        print(kwargs)

    main()



import logging

import click

from ..cli import common_options, logging_options, yes_option
from ..operations import op_queue_context

logger = logging.getLogger(__name__)

try:
    import coverage
    coverage.process_startup()
except ImportError:
    pass


def cli_entry():
    from duqtools import fix_dependencies
    fix_dependencies()
    cli()


@click.group()
def cli(**kwargs):
    """For more information, check out the documentation:

    https://duqtools.readthedocs.io/large_scale_validation
    """


@cli.command('setup')
@click.option(
    '-i',
    '--input',
    'input_file',
    type=click.Path(exists=True),
    help='Input file, i.e. `data.csv` or `runs.yaml`',
    default='data.csv',
)
@click.option(
    '-t',
    '--template',
    'template_file',
    type=click.Path(exists=True),
    help='Template duqtools.yaml',
    default='duqtools.template.yaml',
)
@click.option('--force',
              is_flag=True,
              help='Overwrite existing run config directories')
@common_options(*logging_options, yes_option)
def cli_setup(**kwargs):
    """Set up large scale validation."""
    from .setup import setup
    with op_queue_context():
        setup(**kwargs)


@cli.command('create')
@click.option('--force',
              is_flag=True,
              help='Overwrite existing run directories and IDS data.')
@common_options(*logging_options, yes_option)
def cli_create(**kwargs):
    """Create data sets for large scale validation."""
    from .create import create
    with op_queue_context():
        create(**kwargs)


@cli.command('submit')
@click.option('--force',
              is_flag=True,
              help='Re-submit running or completed jobs.')
@click.option(
    '--schedule',
    is_flag=True,
    help='Schedule and submit jobs automatically. `max_jobs` must be defined.')
@click.option('-j',
              '--max_jobs',
              type=int,
              help='Maximum number of jobs to submit.')
@common_options(*logging_options, yes_option)
def cli_submit(**kwargs):
    """Submit large scale validation runs."""
    from .submit import submit
    with op_queue_context():
        submit(**kwargs)


@cli.command('status')
@click.option('--detailed', is_flag=True, help='Detailed info on progress')
@click.option('--progress', is_flag=True, help='Fancy progress bar')
@common_options(*logging_options)
def cli_status(**kwargs):
    """Check status large scale validation runs."""
    from .status import status
    with op_queue_context():
        status(**kwargs)


@cli.command('merge')
def cli_merge(**kwargs):
    """Merge large scale validation data."""
    from .merge import merge
    with op_queue_context():
        merge(**kwargs)



import itertools
from typing import Optional

from scipy.stats import qmc


def cartesian_product(*iterables, **kwargs):
    """Return cartesian product of input iterables.

    Uses `itertools.product`

    Parameters
    ----------
    *iterables
        Input iterables.

    Returns
    -------
    list[Any]
        List of product of input arguments.
    """
    return list(itertools.product(*iterables))


def _sampler(func, *iterables, n_samples: int, **kwargs):
    """Generic sampler."""
    bounds = tuple(len(iterable) for iterable in iterables)

    sampler = func(d=len(iterables), **kwargs)
    indices = sampler.integers(l_bounds=bounds, n=n_samples)

    samples = []
    for row in indices:
        samples.append(tuple(arg[col] for col, arg in zip(row, iterables)))

    return samples


def latin_hypercube(*iterables,
                    n_samples: int,
                    seed: Optional[int] = None,
                    **kwargs):
    """Sample input iterables using Latin hypercube sampling (LHS).

    Uses `scipy.stats.qmc.LatinHyperCube`.

    Parameters
    ----------
    *iterables
        Iterables to sample from.
    n_samples : int
        Number of samples to return.
    seed : int, optional
        Seed to use for the randomizer

    Returns
    -------
    samples : list[Any]
        List of sampled input arguments.
    """
    return _sampler(qmc.LatinHypercube,
                    *iterables,
                    n_samples=n_samples,
                    seed=seed)


def sobol(*iterables, n_samples: int, seed: Optional[int] = None, **kwargs):
    """Sample input iterables using the Sobol sampling method for generating
    low discrepancy sequences.

    Uses `scipy.stats.qmc.Sobol`.

    Parameters
    ----------
    *iterables
        Iterables to sample from.
    n_samples : int
        Number of samples to return. Note that Sobol sequences lose their
        balance  properties if one uses a sample size that is not a power
        of two.
    seed : int, optional
        Seed to use for the randomizer

    Returns
    -------
    samples : list[Any]
        List of sampled input arguments.
    """
    return _sampler(qmc.Sobol, *iterables, n_samples=n_samples, seed=seed)


def halton(*iterables, n_samples: int, seed: Optional[int] = None, **kwargs):
    """Sample input iterables using the Halton sampling method.

    Uses `scipy.stats.qmc.Halton`.

    Parameters
    ----------
    *iterables
        Iterables to sample from.
    n_samples : int
        Number of samples to return.
    seed : int, optional
        Seed to use for the randomizer

    Returns
    -------
    samples : list[Any]
        List of sampled input arguments.
    """
    return _sampler(qmc.Halton, *iterables, n_samples=n_samples, seed=seed)


_SAMPLERS = {
    'latin-hypercube': latin_hypercube,
    'halton': halton,
    'sobol': sobol,
    'low-discrepancy-sequence': sobol,
    'cartesian-product': cartesian_product,
}


def get_matrix_sampler(name: str):
    """Get sampler.

    Parameters
    ----------
    name : str
        Name of the sampler.

    Returns
    -------
    Callable
        Sampling function.
    """
    return _SAMPLERS[name]



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ..schema import BaseModel, ImasBaseModel

if TYPE_CHECKING:
    from pathlib import Path

    from ..ids import ImasHandle
    from ._job import Job

logger = logging.getLogger(__name__)


class AbstractSystem(ABC, BaseModel):

    @staticmethod
    @abstractmethod
    def get_runs_dir() -> Path:
        """Get the directory where the runs should be stored.

        Parameters
        ----------

        Returns
        -------
        Path
        """
        pass

    @staticmethod
    @abstractmethod
    def write_batchfile(run_dir: Path):
        """Write the batchfile used to submit the job inside run directory
        `run_dir`.

        Parameters
        ----------
        run_dir : Path
            directory of run
        """
        pass

    @staticmethod
    @abstractmethod
    def submit_job(job: Job):
        """submit the job specified by `job` to the system.

        Parameters
        ----------
        job : Job
            job
        """
        pass

    @staticmethod
    @abstractmethod
    def copy_from_template(source_drc: Path, target_drc: Path):
        """copy from template directory `source_drc` to target directory
        `target_drc`

        Parameters
        ----------
        source_drc : Path
            source directory
        target_drc : Path
            target directory
        """
        pass

    @staticmethod
    @abstractmethod
    def imas_from_path(template_drc: Path) -> ImasHandle:
        """It takes a path, and finds out the IMAS entry associated with it.

        Parameters
        ----------
        template_drc : Path
            folder from which to extract IMAS location

        Returns
        -------
        ImasHandle
        """
        pass

    @staticmethod
    @abstractmethod
    def update_imas_locations(run: Path, inp: ImasBaseModel,
                              out: ImasBaseModel):
        """Set the imas entries for the run, both input imas file `in` and
        output imas file `out`.

        Parameters
        ----------
        run : Path
            run directory
        inp : ImasBaseModel
            Imas entry to use as input
        out : ImasBaseModel
            Imas entry to use as output
        """
        pass



from pathlib import Path

import click

from ..config import cfg
from ..models import Job, Locations
from ..status import Status, StatusError


def status(*, progress: bool, detailed: bool, **kwargs):
    """Show status of nested runs.

    Parameters
    ----------
    progress : bool
        Show progress bar.
    detailed : bool
        Show detailed progress for every job.
    """

    cwd = Path.cwd()

    config_files = cwd.glob('**/duqtools.yaml')

    all_jobs: list[Job] = []

    click.echo(Job.status_symbol_help())
    click.echo()

    for config_file in config_files:
        cfg.parse_file(config_file)

        if not cfg.status:
            raise StatusError(
                f'Status field required in config file: {config_file}')

        config_dir = config_file.parent

        jobs = [Job(run.dirname) for run in Locations(config_dir).runs]
        all_jobs.extend(jobs)

        name = config_file.parent.name
        tag = cfg.tag
        status = ''.join(job.status_symbol for job in jobs)

        click.echo(f'{name} ({tag}): {status}')

    click.echo()

    tracker = Status(all_jobs)

    if detailed:
        tracker.detailed_status()
    elif progress:
        tracker.progress_status()
    else:
        tracker.simple_status()



from collections import deque
from pathlib import Path
from typing import Deque

from ..config import cfg
from ..models import Job, Locations
from ..submit import SubmitError, job_scheduler, job_submitter, lockfile_ok, status_file_ok


def submit(*, force, max_jobs, schedule, **kwargs):
    """Submit nested duqtools configs.

    Parameters
    ----------
    force : bool
        Force the submission even in the presence of lockfiles
    max_jobs : int
        Maximum number of jobs to submit at once
    schedule : bool
        Schedule `max_jobs` to run at once, keeps the process alive until
        finished.
    """
    cwd = Path.cwd()

    config_files = cwd.glob('**/duqtools.yaml')

    jobs = []

    for config_file in config_files:
        cfg.parse_file(config_file)

        if not cfg.submit:
            raise SubmitError(
                f'Submit field required in config file: {config_file}')

        config_dir = config_file.parent

        jobs.extend(Job(run.dirname) for run in Locations(config_dir).runs)

    job_queue: Deque[Job] = deque()

    for job in jobs:
        if not status_file_ok(job, force=force):
            continue
        if not lockfile_ok(job, force=force):
            continue
        job_queue.append(job)

    submitter = job_scheduler if schedule else job_submitter

    submitter(job_queue, max_jobs=max_jobs)



from pathlib import Path

import pandas as pd
import streamlit as st
from _shared import default_workdir, get_dataset

from duqtools._plot_utils import alt_errorband_chart, alt_line_chart
from duqtools.config import var_lookup
from duqtools.utils import read_imas_handles_from_file

st.title('Plot IDS')

with st.sidebar:
    st.header('Input data')
    work_dir = st.text_input('Work directory', default_workdir)
    data_file = st.text_input('Data file', 'data.csv')

inp = Path(work_dir) / data_file

if not inp.exists():
    raise ValueError(f'{inp} does not exist!')

handles = read_imas_handles_from_file(inp)
df = pd.DataFrame.from_dict(
    {index: model.dict()
     for index, model in handles.items()}, orient='index')

with st.expander('Click to show runs'):
    st.table(df)

with st.sidebar:
    ids_variables = var_lookup.filter_type('IDS-variable')

    var_names = st.multiselect('Select variable',
                               tuple(ids_variables),
                               default=None)

    st.header('Plotting options')

    show_error_bar = st.checkbox(
        'Show errorbar',
        help=(
            'Show standard deviation band around mean y-value. All '
            'y-values are interpolated to put them on a common basis for x.'),
    )

for variable in (var_lookup[var_name] for var_name in var_names):

    source, time_var, grid_var, data_var = get_dataset(handles, variable)

    st.header(f'{grid_var} vs. {data_var}')

    chart_func = alt_errorband_chart if show_error_bar else alt_line_chart

    if show_error_bar:
        chart = alt_errorband_chart(source, x=grid_var, y=data_var, z=time_var)
    else:
        chart = alt_line_chart(source, x=grid_var, y=data_var, z=time_var)

    st.altair_chart(chart, use_container_width=True)

    if st.button('Save this chart', key=f'download_{grid_var}-{data_var}'):
        fname = f'chart_{grid_var}-{data_var}.html'
        chart.save(fname)
        st.success(f'✔️ Wrote chart to "{fname}"')



from __future__ import annotations

import logging

from .config import var_lookup
from .ids import ImasHandle, merge_data
from .operations import op_queue
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def merge(*, merge_all: bool, target: str, template: str, handles: list[str],
          input_files: list[str], var_names: list[str] | None, force: bool,
          **kwargs):
    """Merge as many data as possible."""
    template = ImasHandle.from_string(template)
    target = ImasHandle.from_string(target)

    handles = [ImasHandle.from_string(handle) for handle in handles]
    for file in input_files:
        handles = handles + list(read_imas_handles_from_file(file).values())

    handles = set(handles)  # Remove duplicate handles

    for handle in handles:
        op_queue.info(description='Source for merge',
                      extra_description=f'{handle}')

    if merge_all:
        ids_variables = tuple(var_lookup.filter_type('IDS-variable').values())

        op_queue.info(description='Merging all known variables')
    else:
        if not var_names or len(var_names) == 0:
            op_queue.add_no_op('No variables specified for merge', 'aborting')
            return

        ids_variables = tuple(var_lookup[name] for name in var_names)
        for variable in ids_variables:
            op_queue.info(description='Variable for merge',
                          extra_description=f'{variable.name}')

    op_queue.add(description='Template for merge',
                 extra_description=f'{template}')

    if target.exists() and not force:
        op_queue.add_no_op(description='Abort merge',
                           extra_description=f'{target} already exists, '
                           'use --force to overwrite')
        return

    template.copy_data_to(target)

    merge_data(handles, target, ids_variables)



from typing import Optional

from pydantic import Field

from ._basemodel import BaseModel
from ._description_helpers import formatter as f


class DataLocation(BaseModel):
    """Specification for the data generated by the create step.

    When setting up a sequence of UQ runs, duqtools reads the source data from
    the template. For each individual UQ run needs, two locations must be
    defined.
        1. The location of the input data. This is where duqtools stores
        the modified source data.
        2. The location of the output data. The modelling software must know
        in advance where to store the results of the simulation.

    Input data are defined by `run_in_start_at`, and output data by
    `run_out_start_at`. A sequence is generated starting from these numbers.

    For example, with `run_in_start_at`: 7000 and `run_out_start_at`: 8000,
    the generated input stored at run number 7000 would correspond to output
    8000, 7001 to 8001, 7002 to 8002, etc.

    Note that these sequences may overlap with existing data sets. Duqtools
    will stop if it detects that data will be overwritten.
    """

    user: Optional[str] = Field(
        description='Username for the IMAS database to use,'
        ' defaults to current user')

    imasdb: str = Field(description='IMAS database or machine name.')

    run_in_start_at: int = Field(description=f("""
            The sequence of input data files start with this run number.
            """))

    run_out_start_at: int = Field(description=f("""
            The sequence of output data files start with this run number.
            """))



from os import getenv
from pathlib import Path
from typing import Optional

from ..config import cfg
from ..schema.runs import Run, Runs


class Locations:
    """Class that knows about locations within a duqtools config directory.

    Defaults to local working directory, but can be initialized with a
    known config directory.
    """

    def __init__(self, parent_dir: Optional[Path] = None):
        if not parent_dir:
            parent_dir = Path.cwd()

        self.parent_dir = parent_dir

    @property
    def data_csv(self):
        """Location of data.csv."""
        return self.parent_dir / 'data.csv'

    @property
    def runs_yaml(self):
        """Location of runs.yaml."""
        return self.parent_dir / 'runs.yaml'

    @property
    def runs_yaml_old(self):
        """Location of runs.yaml.old."""
        return self.parent_dir / 'runs.yaml.old'

    @property
    def runs(self) -> list[Run]:
        """Get a list of the runs currently created from this config."""
        runs_yaml = self.runs_yaml

        if not runs_yaml.exists():
            raise OSError(f'Cannot find {runs_yaml}.')

        return Runs.parse_file(runs_yaml)

    @property
    def jruns_path(self) -> Path:
        """return the Path specified in the create->jruns config variable, or,
        if that is empty, the `$JRUNS` environment variable, or, if `$JRUNS`
        does not exists, return the current directory `./`.

        Returns
        -------
        Path
        """

        if cfg.create.jruns:  # type: ignore
            return cfg.create.jruns  # type: ignore
        elif getenv('JRUNS'):
            return Path(getenv('JRUNS'))  # type: ignore
        else:
            return Path()



import logging
import time
from collections import deque
from itertools import cycle
from pathlib import Path
from typing import Deque, Sequence

from ._logging_utils import duqlog_screen
from .config import cfg
from .models import Job, Locations
from .operations import add_to_op_queue, op_queue
from .system import get_system

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


class SubmitError(Exception):
    ...


def Spinner(frames='⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏'):
    """Simple spinner animation."""
    yield from cycle(frames)


@add_to_op_queue('Submitting', '{job}')
def _submit_job(job: Job, *, delay: float = 0):
    """
    Parameters
    ----------
    delay : float
        Add a delay to avoid all jobs starting at the same time.
        This causes issues with slurm, for example.
    """
    if delay:
        time.sleep(delay)
    job.submit()


def job_submitter(jobs: Sequence[Job], *, max_jobs):
    for n, job in enumerate(jobs):
        if max_jobs and (n >= max_jobs):
            info(f'Max jobs ({max_jobs}) reached.')
            break

        _submit_job(job, delay=0.1)


@add_to_op_queue('Start job scheduler')
def job_scheduler(queue: Deque[Job], max_jobs=10):
    interval = 1.0

    s = Spinner()

    tasks: Deque[Job] = deque()
    completed: Deque[Job] = deque()

    while tasks or queue:
        while queue and len(tasks) < max_jobs:
            job = queue.popleft()
            task = job.start()
            tasks.append(task)
            # starting jobs at the same time causes issues
            time.sleep(0.1)

        time.sleep(interval)
        task = tasks.popleft()
        try:
            next(task)  # Run to the next yield
            tasks.append(task)  # Reschedule
        except StopIteration:
            completed.append(task)

        print(
            f' {next(s)} Running: {len(tasks)},'
            f' queue: {len(queue)}, completed: {len(completed)}',
            end='\033[K\r',
        )


def job_array_submitter(jobs: Sequence[Job], *, max_jobs):
    if len(jobs) == 0:
        duqlog_screen.error('No jobs to submit, not creating array ...')
        return

    for job in jobs:
        op_queue.add(action=lambda: None,
                     description='Adding to array',
                     extra_description=f'{job}')

    get_system().submit_array(jobs)


def submission_script_ok(job):
    submission_script = job.submit_script
    if not submission_script.is_file():
        info('Did not found submission script %s ; Skipping directory...',
             submission_script)
        return False

    return True


def status_file_ok(job, *, force):
    status_file = job.status_file
    if job.has_status and not force:
        if not status_file.is_file():
            logger.warning('Status file %s is not a file', status_file)
        with open(status_file) as f:
            info('Status of %s: %s. To rerun enable the --force flag',
                 status_file, f.read())
        op_queue.add_no_op(
            description='Not Submitting',
            extra_description=f'{job} (reason: status file exists)')
        return False

    return True


def lockfile_ok(job, *, force):
    lockfile = job.lockfile
    if lockfile.exists() and not force:
        op_queue.add_no_op(
            description='Not Submitting',
            extra_description=f'{job} (reason: {lockfile} exists)')
        return False

    return True


def get_resubmit_jobs(resubmit_names: Sequence[Path]) -> list[Job]:
    """get_resubmit_jobs.

    Parameters
    ----------
    resubmit_names : Sequence[Path]
        The names (short or full path) of the jobs that need to be resubmitted

    Returns
    -------
    list[Job]
    """
    jobs: list[Job] = []
    run_dict = {run.shortname: run for run in Locations().runs}
    for name in resubmit_names:
        if name in run_dict:  # check for shortname
            jobs.append(Job(run_dict[name].dirname))
        else:
            # It's not a short name, use the argument as the full path to the job
            jobs.append(Job(Path(name)))
    return jobs


def submit(*,
           force: bool,
           max_jobs: int,
           schedule: bool,
           array: bool,
           resubmit: Sequence[Path] = (),
           **kwargs):
    """submit. Function which implements the functionality to submit jobs to
    the cluster.

    Parameters
    ----------
    force : bool
        Force the submission even in the presence of lockfiles
    max_jobs : int
        Maximum number of jobs to submit at once
    schedule : bool
        Schedule `max_jobs` to run at once, keeps the process alive until
        finished.
    array : bool
        Submit the jobs as a single array
    resubmit : Sequence[Path]
        If any jobs need to be resubmitted, this is has a nonzero length, and
        contains a Sequence of Paths which either are the full Path to the run
        that needs to be resubmitted, or the shortname
    """
    if not cfg.submit:
        raise SubmitError('Submit field required in config file')

    debug('Submit config: %s', cfg.submit)

    if resubmit:
        jobs = get_resubmit_jobs(resubmit)
        force = True
    else:
        jobs = [Job(run.dirname) for run in Locations().runs]

    debug('Case directories: %s', jobs)

    job_queue: Deque[Job] = deque()

    for job in jobs:
        if not status_file_ok(job, force=force):
            continue
        if not lockfile_ok(job, force=force):
            continue
        job_queue.append(job)

    if array and Path('./duqtools_slurm_array.sh').exists() and not force:
        op_queue.add_no_op(
            description='Not Creating Array',
            extra_description='(reason: duqtools_slurm_array.sh exists)')
        op_queue.add_no_op(description='Not Submitting Array',
                           extra_description='use --force to override')
        return

    submitter = job_scheduler if schedule else job_submitter
    submitter = job_array_submitter if array else submitter
    submitter(job_queue, max_jobs=max_jobs)



import click
from pydantic import ValidationError

from duqtools.config import cfg, var_lookup

cs = click.style

ST_ITEMS = {'fg': 'green', 'bold': True}
ST_HEADER = {'fg': 'red', 'bold': True}
ST_INFO = {'fg': 'white', 'bold': False}


def list_group(group: list, extra_variables: dict):
    group = sorted(group, key=lambda var: var.name)

    for var in group:
        star = cs('*', **ST_HEADER) if var.name in extra_variables else ''

        name = cs(f'{var.name}', **ST_ITEMS)

        try:
            sub = cs(f'({var.path})', **ST_INFO)
        except AttributeError:
            sub = cs(f'({var.type})', **ST_INFO)

        click.echo(f'    - {star}{name} {sub}')


def list_variables(*, config, **kwargs):
    """List variables in `variables.yaml` and config/`duqtools.yaml` if
    present.

    Parameters
    ----------
    config : str
        Name of the config file to use
    **kwargs
        Unused.
    """
    try:
        cfg.parse_file(config)
    except FileNotFoundError:
        print(f'Could not find: {config}')
    except ValidationError as e:
        exit(e)
        print(f'*: defined by {config}')
    finally:
        extra_variables = cfg.extra_variables.to_variable_dict(
        ) if cfg.extra_variables else {}

    grouped_ids_vars = var_lookup.groupby_ids()

    for root_ids, group in grouped_ids_vars.items():
        click.secho(f'\nIDS-variable - {root_ids}:', **ST_HEADER)
        list_group(group, extra_variables)

    grouped_other_vars = var_lookup.groupby_type()
    grouped_other_vars.pop('IDS-variable')

    for var_type, group in grouped_other_vars.items():
        click.secho(f'\n{var_type}:', **ST_HEADER)
        list_group(group, extra_variables)



import json

from ..schema import JettoVar, JsetField, NamelistField


def jettovar_to_json(variable: JettoVar):
    jsetfields = []
    nmlfields = []
    for field in variable.keys:
        if isinstance(field, JsetField):
            jsetfields.append(field)
        elif isinstance(field, NamelistField):
            nmlfields.append(field)
        else:
            raise NotImplementedError(f'unknown Jetto field type {field}')

    if len(nmlfields) > 1:
        raise RuntimeError(
            f'jetto_tools only support a single nml entry: {nmlfields}')

    var_dict = {'type': variable.type, 'dimension': 'scalar'}
    if var_dict['type'] == 'float':
        var_dict['type'] = 'real'

    if len(jsetfields) > 0:
        var_dict['jset_id'] = jsetfields[0].field
    if len(jsetfields) > 1:
        var_dict['jset_flex_id'] = []  # type: ignore
        for field in jsetfields[1:]:
            var_dict['jset_flex_id'].append(field.field)  # type: ignore

    if len(nmlfields) == 1:
        var_dict['nml_id'] = {  # type: ignore
            'namelist': nmlfields[0].section,
            'field': nmlfields[0].field
        }

    return json.dumps({variable.name: var_dict})



from __future__ import annotations

import logging
import operator
import os
import sys
from collections import UserDict
from pathlib import Path, PosixPath
from typing import Hashable, Sequence

from ..schema import IDSVariableModel
from ..schema.variables import VariableConfigModel
from ..utils import groupby

if sys.version_info < (3, 10):
    from importlib_resources import files
else:
    from importlib.resources import files

logger = logging.getLogger(__name__)

VAR_ENV = 'DUQTOOLS_VARDEF'
USER_CONFIG_HOME = Path.home() / '.config'
LOCAL_DIR = Path('.').absolute()
DUQTOOLS_DIR = 'duqtools'
VAR_FILENAME = 'variables.yaml'
VAR_FILENAME_GLOB = 'variables*.yaml'


class VarLookup(UserDict):
    _prefix = '$'
    """Variable lookup table.

    Subclasses `UserDict` to embed some commonly used operations, like
    grouping and filtering.
    """
    _ids_variable_key = 'IDS-variable'

    def __getitem__(self, key: str) -> IDSVariableModel:
        return self.data[self.normalize(key)]

    def normalize(self, *keys: str) -> str | tuple[str, ...]:
        """Normalize variable names (remove `$`)."""
        keys = tuple(key.lstrip(self._prefix) for key in keys)
        if len(keys) == 1:
            return keys[0]
        return keys

    def filter_type(self, type: str, *, invert: bool = False) -> VarLookup:
        """Filter all entries of given type."""
        cmp = operator.ne if invert else operator.eq
        return VarLookup({k: v for k, v in self.items() if cmp(v.type, type)})

    def groupby_type(self) -> dict[Hashable, list[IDSVariableModel]]:
        """Group entries by type."""
        grouped_ids_vars = groupby(self.values(), keyfunc=lambda var: var.type)
        return grouped_ids_vars

    def filter_ids(self, ids: str) -> VarLookup:
        """Filter all entries of given IDS."""
        ids_vars = self.filter_type(self._ids_variable_key)

        return VarLookup({k: v for k, v in ids_vars.items() if v.ids == ids})

    def groupby_ids(self) -> dict[Hashable, list[IDSVariableModel]]:
        """Group entries by IDS."""
        ids_vars = self.filter_type(self._ids_variable_key).values()

        grouped_ids_vars = groupby(ids_vars, keyfunc=lambda var: var.ids)
        return grouped_ids_vars


class VariableConfigLoader:
    def __init__(self):
        self.paths = self.get_config_path()

    def load(self) -> VarLookup:
        """Load the variables config."""
        var_lookup = VarLookup()

        for path in self.paths:
            logger.debug(f'Loading variables from: {path}')

            var_config = VariableConfigModel.parse_file(path)
            var_lookup.update(var_config.to_variable_dict())

        return var_lookup

    def get_config_path(self) -> tuple[Path, ...]:
        """Try to get the config file with variable definitions.

        Search order:
        1. environment variable
        (2. local directory, not sure if this should be implemented)
        3. config home (first $XDG_CONFIG_HOME/duqtools then `$HOME/.config/duqtools`)
        4. fall back to variable definitions in package
        """
        for paths in (
                self._get_paths_from_environment_variable(),
                self._get_paths_from_config_home(),
                self._get_paths_local_directory(),
        ):
            if paths:
                return paths

        return self._get_paths_fallback()

    def _get_paths_from_environment_variable(self) -> tuple[Path, ...] | None:
        env = os.environ.get(VAR_ENV)
        if env:
            path = Path(env)
            drc = path.parent

            if not drc.exists():
                raise OSError(f'{path} defined by ${VAR_ENV} does not exist!')

            return tuple(drc.glob(path.name))

        return None

    def _get_paths_local_directory(self) -> tuple[Path, ...] | None:
        return None  # Not implemented

    def _get_paths_from_config_home(self) -> tuple[Path, ...] | None:
        config_home = os.environ.get('XDG_CONFIG_HOME', USER_CONFIG_HOME)

        drc = Path(config_home) / DUQTOOLS_DIR
        if drc.exists():
            return tuple(drc.glob(VAR_FILENAME_GLOB))

        return None

    def _get_paths_fallback(self) -> tuple[Path, ...]:
        module = files('duqtools.data')
        assert module.is_dir()
        drc: PosixPath = module._paths[0]  # type: ignore
        return tuple(drc.glob(VAR_FILENAME_GLOB))


def lookup_vars(
        variables: Sequence[(str | IDSVariableModel)]
) -> list[IDSVariableModel]:
    """Helper function to look up a bunch of variables.

    If str, look up the variable from the `var_lookup`. Else, check if
    the variable is an `IDSVariableModel`.
    """
    var_models = []
    for var in variables:
        if isinstance(var, str):
            var = var_lookup[var]
        if not isinstance(var, IDSVariableModel):
            raise ValueError(f'Cannot lookup variable with type {type(var)}')
        var_models.append(var)
    return var_models


var_lookup = VariableConfigLoader().load()



import sys
from pathlib import Path
from typing import Sequence

import streamlit as st
import xarray as xr

from duqtools.api import ImasHandle, standardize_grid_and_time
from duqtools.config import var_lookup

try:
    default_workdir = sys.argv[1]
except IndexError:
    default_workdir = str(Path.cwd())

PREFIX = 'profiles_1d/*'
PREFIX0 = 'profiles_1d/0'


@st.experimental_memo
def get_ids_options():
    return tuple(var_lookup.filter_type('IDS-variable').keys())


@st.experimental_memo
def get_var_options(ids):
    return list(var_lookup.filter_ids(ids).values())


@st.experimental_memo
def get_variables(*, ids: str, x_key: str, y_keys: Sequence[str]):
    return {
        'time_var': 'time',
        'grid_var': x_key,
        'data_vars': y_keys,
    }


@st.experimental_memo
def _get_dataset(handles, variable):
    data_var = variable['name']
    time_var = variable['dims'][0]
    grid_var = variable['dims'][1]

    datasets = []

    for name, handle in handles.items():
        handle = ImasHandle(**handle)
        ds = handle.get_variables(variables=(time_var, grid_var, data_var))
        datasets.append(ds)

    grid_var_norm = var_lookup.normalize(grid_var)
    time_var_norm = var_lookup.normalize(time_var)

    datasets = standardize_grid_and_time(
        datasets,
        grid_var=grid_var_norm,
        time_var=time_var_norm,
    )

    dataset = xr.concat(datasets, 'run')
    dataset['run'] = list(handles.keys())

    return dataset, time_var_norm, grid_var_norm, data_var


def get_dataset(handles, variable):
    """Convert to hashable types before calling `_get_dataset`."""
    handles = {name: handle.dict() for name, handle in handles.items()}
    variable = variable.dict()

    return _get_dataset(handles, variable)



from ._basemodel import BaseModel
from ._dimensions import IDSOperation, IDSOperationDim, JettoOperation, OperationDim
from ._imas import ImasBaseModel
from ._jetto import JettoField, JettoVar, JsetField, NamelistField
from ._ranges import ARange, LinSpace
from ._variable import IDSVariableModel, JettoVariableModel

__all__ = [
    'ARange',
    'BaseModel',
    'IDSOperation',
    'IDSOperationDim',
    'IDSVariableModel',
    'ImasBaseModel',
    'JettoField',
    'JettoOperation',
    'JettoVar',
    'JettoVariableModel',
    'JsetField',
    'LinSpace',
    'NamelistField',
    'OperationDim',
]
